import numpy as np
from scipy import special
from astropy.coordinates import SkyCoord
from astropy import units as u

try:
    import bottleneck as bn
    MEDIAN = bn.median
    NANMEDIAN = bn.nanmedian
    MEAN = bn.nanmean
    STD = bn.nanstd
except ImportError:
    MEDIAN = np.median
    NANMEDIAN = np.nanmedian
    MEAN = np.mean
    STD = np.std

def _get_start_stop(arr,arr_in):
    """
    arr_in: array
       sorted
    """
    if not np.all(arr_in[:-1] <= arr_in[1:]):
        raise('arr_in should be sorted')
    inds_left= np.searchsorted(arr_in, arr, side='left',)
    inds_right=np.searchsorted(arr_in, arr, side='right',)
    return inds_left,inds_right

def conv_fun(dis, beamsize=2.9/60, kernel='bessel_gaussian'):
    """
    convolving functions in Mangum 2007
    
    dis: degree
    beamsize: degree
    kernel: gaussian, bessel_gaussian, sinc_gaussian
    
    """
    if kernel == 'gaussian':
        return np.exp(-(dis/(beamsize/3))**2)
    
    a = 1.55 * beamsize / 3.
    b = 2.52 * beamsize / 3.
    x = np.pi*dis/a 
    wei = np.empty_like(dis)
    not_zero = dis != 0
    wei[~not_zero] = 1
    if kernel == 'bessel_gaussian':
        wei[not_zero] = 2*special.j1(x[not_zero])/(x[not_zero])*np.exp(-(dis[not_zero]/b)**2)
    elif kernel == 'sinc_gaussian':
        wei[not_zero] = np.sin(x[not_zero])/(x[not_zero])*np.exp(-(dis[not_zero]/b)**2)
    else:
        raise(ValueError(f'kernel \"{kernel}\" is not supported'))
    return wei


def pixel_spec(spec, dis, *, wi=None, method='bessel_gaussian', sigma=1.275088/60, beamsize=2.9/60, statistic='median', frac_finite_min=1):
    """
    -----------------
    spec: flux
    dis: degree
    wi: initial weight, only some method support it.
    method: str; 'bessel_gaussian', 'gaussian', 'sinc_gaussian', 'reweight', 'mean', 'median'
    sigma: degree, used in 'reweight'
    statistic: str; median or mean, default is median; used in 'reweight'
    frac_finite_min: if the number of ``finite value`` in a channel is zero or smaller than ``frac_finite_min*len(spec)``, 
                     the output value in it will be set as np.nan
    """
    
    spec = spec.astype('float64')
    
    if frac_finite_min < 1:
        _MEAN = np.nanmean
        _MEDIAN = np.nanmedian
        _SUM = np.nansum
        is_finite = np.isfinite(spec)
    else:
        _MEAN = np.mean
        _MEDIAN = np.median
        _SUM = np.sum
    
    if method == 'bessel_gaussian' or method == 'gaussian' or method =='sinc_gaussian':
        wei = conv_fun(dis, beamsize, kernel=method)
        if wi is not None:
            wei *= wi
        wei = wei.reshape((-1,)+(1,)*(spec.ndim-1))
        if frac_finite_min < 1:
            s = _SUM(spec*wei, axis=0)/_SUM(wei*is_finite, axis=0)
        else:
            s = _SUM(spec*wei, axis=0)/_SUM(wei)
            
    elif method == "reweight":
        #Barnes el. al. 2001, MNRAS 322, 486 https://ui.adsabs.harvard.edu/abs/2001MNRAS.322..486B/abstract
        if statistic == 'median':
            wei_m = _MEDIAN(np.exp(- (dis/sigma)**2/2))
            if wi is not None:
                wei_m *= wi
            s = _MEDIAN(spec, axis=0)/wei_m
        elif statistic == 'mean':
            wei_m= _MEAN(np.exp(- (dis/sigma)**2/2))
            if wi is not None:
                wei_m *= wi
            s = _MEAN(spec, axis=0)/wei_m
    elif method == 'mean':
        s = _MEAN(spec, axis=0)
    elif method == 'median':
        s = _MEDIAN(spec, axis=0)
    else:
        raise(ValueError('method'))
        
    if frac_finite_min < 1:
        s[np.sum(is_finite, axis=0) < len(spec)*frac_finite_min] = np.nan
    
    return s

def gridding(ra, dec, spectra, ra_grid, dec_grid, *, wi=None, r=1.5/60, **kwargs):
    """
    ra, dec: array, shape (m,); degree
        The ra dec of observed spectra; degree
    spectra: array, shape (m,n)
        flux
    ra_grid, dec_grid: array, shape (x,y) or (x,); degree
        The center of the grid 
    r: scalar; degree
       The spectra separated from the center of the grid less than r will be considered. 
    ------------
    other parameters
      parameters in function pixel_spec
    method: str; 'bessel_gaussian', 'gaussian', 'sinc_gaussian', 'reweight', 'mean', 'median'
    sigma: degree, used in 'reweight'
    statistic: str; median or mean, default is median; used in 'reweight'
    """
    if np.isscalar(ra) or np.isscalar(dec) or np.isscalar(ra_grid) or np.isscalar(dec_grid):
        raise ValueError('One of the inputs is a scalar.')
    cata = SkyCoord(ra,dec,unit=(u.degree, u.degree))
    grid= SkyCoord(ra_grid,dec_grid, unit=(u.degree, u.degree))
    grid_ori_ndim=grid.ndim
    if grid_ori_ndim==1:
        grid=grid[:,None]
    grid_f= grid.flatten()
    #find the spec in r arcmin
    ind_g, ind_cata, d2d, d3d=cata.search_around_sky(grid_f,r*u.degree)
    ind_g_uni=np.unique(ind_g)# index in grid flatten
    start, stop= _get_start_stop(ind_g_uni,ind_g)
    
    out= np.full(spectra.shape[1:2] + grid.shape, np.nan)
    nums= np.zeros(grid.shape)
    for i,start_, stop_ in zip(ind_g_uni,start, stop):
        dis= d2d[start_ : stop_ ] #distance of spec from the center of grid
        ind_use_= ind_cata[start_ : stop_ ] # index in cata, ra, dec, spectra
        spec_= spectra[ind_use_]
        wi_ = wi[ind_use_] if wi is not None else None
        if len(spec_) == 0:
            continue
        m,n=i//grid.shape[1], i%grid.shape[1] #index in grid before flatten
        nums[m,n]=len(dis)
        out[:,m,n]= pixel_spec(spec_, dis.degree, wi=wi_, **kwargs)
    
    if grid_ori_ndim==1:
        out= out[:,:,0]
        nums= nums[:,0]
    return out, nums



__all__ = ['replace_spec', 'save_rep_fixed', 'repalce_near', 'replace_margin_side', 'get_margin_num', 'replace_rfi_lower', 'replace_rfi',
           'find_loc', 'SW_FFT', 'get_sw_conf', 'fit_sw_fft',]

# author: Xu Chen, Li Fujia, 2021.06
# code: Xu Chen

import numpy as np
#import numpy.fft as fft
from scipy import fft
from astropy import log
from copy import deepcopy
from tqdm import tqdm
from matplotlib import pyplot as plt

# Cell
def replace_spec(tn, spec, freq, exceed_use, restrict_use, rfi_width_lim, ext_sec, ext, STD, MAX=None,
                 test=False, sm = None, fill = 'original', shift = 0,):
    """
    replace mw, rfi or others by near ripple section in one spectrum
    exceed_use: spec > thr | rfi
    restrict_use: restrict replace area in exceed_use
    ext: extend freq range to replace (channel number)
    rfi_width_lim: rfi should contain more channels than limit
    ext_sec: extend channel number of start and end of each section
    """
    if MAX is None:
        MAX = np.nanmax(spec)
    from .markRFI import rms, get_startend
    from ..utils.misc import smooth1d
    if np.sum(exceed_use) == 0:
        return spec

    start, end = get_startend(exceed_use, rfi_width_lim, ext_sec)
    if len(start) == 0:
#         print(f"tn={tn} raised a warning")
        return spec
    N = len(freq)
    newspec = deepcopy(spec)
    newspec_ = deepcopy(spec)
    if test:
        rep_from = np.zeros_like(spec)
        rep_to = np.zeros_like(spec, dtype='bool')

    for s, e in zip(start, end):
#         print(freq[s],freq[e])
        if fill == 'zero':
            newspec[s:e] = 0
            if test: rep_to[s:e] = 1
        else:
            # used to find two min values as sin valleys
            spec1 = np.zeros_like(freq) + MAX * 20
            s0 = s - ext
            s0 = max(s0, 0)
            e0 = e + ext
            e0 = min(e0, N)
    #         print(freq[s0],freq[e0])
            spec1[s0:e0] = sm[s0:e0]

            spec1l = deepcopy(spec1)
            spec1r = deepcopy(spec1)

            spec1l[s:] = MAX * 20
            l1 = np.argmin(spec1l) if s > 0 else 0
            spec1r[:e] = MAX * 20
            r1 = np.argmin(spec1r) if e < N else N

            if test:
                rep_to[l1:r1] = True
            # [l1:r1] will be replaced, length = L
            # print(l1,r1)
            L = r1 - l1
            if l1 - L - int(L * shift) < 0:
                direc = 'right'
            elif r1 + L + int(L * shift)> N - 1:
                direc = 'left'
            else:
                rmsl = rms(sm[l1 - L:l1], freq[l1 - L:l1])
                rmsr = rms(sm[r1:r1 + L], freq[r1:r1 + L])
                if rmsl < rmsr:
                    direc = 'left'
                else:
                    direc = 'right'

            if direc == 'left':
                s1 = l1 - L     - int(L * shift)
                e1 = l1         - int(L * shift)
            elif direc == 'right':
                s1 = r1         + int(L * shift)
                e1 = r1 + L     + int(L * shift)

            # print(direc,s1,e1)
            # [s1:e1] will be copied into [l1:r1]
            # fill what kinds of noise is not important, because they are high frequency compoents.
            if fill == 'original':
                ripple = spec[s1:e1]
            elif fill == 'smooth':
                ripple = sm[s1:e1]
            elif fill == 'artificial':
                ripple = sm[s1:e1] + np.random.normal(scale=STD, size=int(e1-s1))
            if test: rep_from[s1:e1] = 1
            try:
                newspec_[l1:r1] = ripple
                newspec[s:e] = newspec_[s:e] # only s to e will be changed
#                 print(s, e, l1, r1)
            except ValueError:
                log.info(f"tn = {tn} has a ValueError, will replace with noise")
                # import traceback
                # traceback.print_exc()
                newspec[s:e] = np.random.normal(scale=STD, size=int(e-s)) + np.nanmedian(spec)
    # restrict bound
    if not restrict_use.all(): newspec[~restrict_use] = spec[~restrict_use]
    if test:
        return newspec, rep_from, rep_to
    else:
        return newspec
    
def save_rep_fixed(sm, freq, exceed_use, rfi_width_lim, ext_sec, step, freq_d, freq_u):
    """
    save the extended replaced area, fixed freq step
    """
    from .markRFI import get_startend
    
    is_excluded = np.zeros_like(sm, dtype='bool') 
    if np.sum(exceed_use) == 0:
        return is_excluded

    start, end = get_startend(exceed_use, rfi_width_lim, ext_sec,)
    if len(start) == 0:
        return is_excluded
    
    N = len(freq)
    for s, e in zip(start, end):
        if e > freq_d and s < freq_u:
            if step > 0:
                s0 = s - step
                s0 = max(s0, 0)
                e0 = e + step
                e0 = min(e0, N)
            else:
                s0 = s 
                e0 = e
            is_excluded[s0:e0] = True
        
    return is_excluded
    
def save_rep_2sides(sm, freq, exceed_use, rfi_width_lim, ext_sec, ext, thr, freq_d, freq_u):
    """
    save the extended replaced area, from peak to 2 sides
    """
    from .markRFI import get_startend, find_edge_2sides

    is_excluded = np.zeros_like(sm, dtype='bool')
    if np.sum(exceed_use) == 0:
        return is_excluded

    start, end = get_startend(exceed_use, rfi_width_lim, ext_sec,)
    if len(start) == 0:
        return is_excluded

    N = len(freq)
    for s, e in zip(start, end):
        if e > freq_d and s < freq_u:
            s0 = s - ext * 5
            s0 = max(s0, 0)
            e0 = e + ext * 5
            e0 = min(e0, N)
            peak_position = freq[s:e][np.argmax(sm[s:e])]
#             if e0 < N: print(s,e,s0,e0,freq[s],freq[e],freq[s0],freq[e0], peak_position)
            mask_frange = find_edge_2sides(sm[s0:e0],freq[s0:e0],peak_position = peak_position,
                    small_rfi_times = 1, step=1*ext*fdelta, rms_thresh=thr,Print=False)
            if not np.isnan(mask_frange[0]):
                dist = np.max(np.abs(mask_frange - peak_position))
                mask_frange = peak_position + np.array([-1,1]) * dist
#                 print(mask_frange)
                is_excluded[(freq>=mask_frange[0])&(freq<=mask_frange[1])] = True
        
    return is_excluded

# Cell
def repalce_near(data_in, freq, time_rfi, mw_use=None, times_thr=None, times_s_thr=None, times_s_thr2=None,
                 rms_sigma=None,  ext_freq=None, rms_frange=None, rfi_width_lim=None,
                 ext_sec=None, data_find = None, data_trough = None, data_restrict = None,
                 save_is_excluded = False,ex_step = 0, verbose = True, **kwargs):
    """
    replace mw, rfi or others by near ripple section

    times_thr: above ~ times of rms will be set noise (unsmooth)
    times_s_thr: above ~ times of rms will be replaced (smoothed once)
    times_s_thr2: above ~ times of rms will be replaced (smooth twice)
    ext_freq: extend freq range to replace (mhz)
    rfi_width_lim: rfi should contain more channels than limit
    ext_sec: extend channel number of start and end of each section
    """
    global fdelta
    fdelta = np.abs(freq[1]-freq[0])
    ext = int(np.around(ext_freq / fdelta))
    
    data = deepcopy(data_in)
    
    from .markRFI import real_std, real_rms, std, rms
    if time_rfi is not None:
        whole_rfi = np.all(time_rfi, axis=1)
        is_rfi_num = np.arange(data.shape[0])[whole_rfi]
        not_rfi_num = np.arange(data.shape[0])[~whole_rfi]
    else:
        not_rfi_num = np.arange(data.shape[0])
        time_rfi = np.full(data.shape, False)

    if mw_use is None:
        mw_use = np.full(freq.shape, False)

    save_is_excluded = True if save_is_excluded and (times_s_thr2 is not None) else False

    data_rep = deepcopy(data)
    if data_find is None: data_find = data
    if times_s_thr2 is None: times_s_thr2 = times_s_thr
    if data_trough is None: data_trough = data_find
    if data_restrict is None:
        RESTRICT = False
    else:
        RESTRICT = False if data_restrict.shape != data.shape else True
    
    if save_is_excluded:
        is_excluded = np.zeros_like(data,dtype=bool)        
        Lfreq = len(freq)
        freq_d = int(Lfreq * 0.1)
        freq_u = int(Lfreq * 0.9)
        save_ex_step = int(np.around(ex_step / fdelta))
    else:
        is_excluded = np.array([None])

    from ..utils.misc import smooth1d
    MAX = np.nanmax(data)*20
    
    if np.sum(np.isnan(data)) > 0:
        data[np.isnan(data)] = MAX
        
    if verbose:
        iter_ = tqdm(range(data.shape[0]), desc='CPU 0: ', mininterval=2)
    else:
        iter_ = range(data.shape[0])
    
    for tn in iter_:
        if tn in not_rfi_num:
            spec = deepcopy(data[tn, :])
            include = mw_use #| time_rfi[tn]

            find = ex_sm = data_find[tn]
            # find used to define replace area
            RMS = rms(find, freq, rms_vrange=rms_frange)
            thr_s = RMS*times_s_thr2
            exceed_use = (np.abs(find) > thr_s) | time_rfi[tn]
            if np.sum(include) > 0:
                spec[include] = MAX * 20
                exceed_use |= include
            
            if RESTRICT:
                restrict = ex_sm = data_restrict[tn]
                # restrict used to restrict replace area in one spectra
                restrict_use = (restrict > thr_s) | time_rfi[tn]
            else:
                restrict_use = np.full(freq.shape, True)
            
            if save_is_excluded:
                # save sources or RFI positions
                is_excluded[tn,:] = save_rep_fixed(ex_sm, freq, exceed_use, rfi_width_lim, 
                                             ext_sec, save_ex_step, freq_d,freq_u)
#                 is_excluded[tn,:] = save_rep_2sides(ex_sm, freq, exceed_use, rfi_width_lim, 
#                                              ext_sec, ext, thr_s, freq_d,freq_u)

            # trough used to find ripples valleys
            trough = data_trough[tn]
            STD = real_std(spec, freq, sigma = rms_sigma, vrange = rms_frange)
            newspec = replace_spec(tn, spec, freq, exceed_use, restrict_use, rfi_width_lim, ext_sec,
                                   ext, STD, MAX, sm = trough, **kwargs)

            RMS = real_rms(spec, freq, sigma=rms_sigma, rms_vrange=rms_frange)
            strange = np.where(np.abs(newspec) > times_thr * RMS)[0]
            newspec[strange] = np.random.normal(scale=RMS, size=len(strange))

            data_rep[tn, :] = newspec

    if save_is_excluded:
        # two edges do not mask 
        is_excluded[:,:freq_d] = False
        is_excluded[:,freq_u:] = False
        
    return data_rep, is_excluded

# Cell

def replace_margin_side(data, freq, is_rfi, mw_use, margin_width=20, ext_freq=1.3, side='both'):
    margin = int(len(freq)/(freq[-1]-freq[0])*margin_width)

    N = data.shape[0]
    MAX = np.nanmax(data)*20
    pads = np.full((N, margin), MAX)

    fdelta = freq[1]-freq[0]
    fext1 = np.arange(freq[0]-margin*fdelta, freq[0], fdelta)
    fext2 = np.arange(freq[-1], freq[-1]+margin*fdelta, fdelta)
    print(f"Margin from {side} side(s) ...")

    if side == 'both':
        data_ = np.hstack((pads, data, pads))
        freq_ = np.hstack((fext1, freq, fext2))
    elif side == 'left':
        data_ = np.hstack((pads, data))
        freq_ = np.hstack((fext1, freq))
    elif side == 'right':
        data_ = np.hstack((data, pads))
        freq_ = np.hstack((freq, fext2))

    ext = int(np.around(ext_freq / fdelta))

    from hifas.ripple.markRFI import real_std
    whole_rfi = np.all(is_rfi, axis=1)
    is_rfi_num = np.arange(data.shape[0])[whole_rfi]

    for ti in tqdm(range(N)):
        if ti not in is_rfi_num:
            spec_ = data_[ti]
            exceed_use = (spec_ == np.inf)
            STD = real_std(spec_, freq_, sigma=6, vrange=[1390, 1400])
            restrict_use = np.full(freq.shape, False)
            newspec = replace_spec(ti, spec_, freq_, exceed_use, restrict_use,
                                   rfi_width_lim=0, ext_sec=1, ext=ext, STD = STD, MAX=MAX)
            newspec[-1] = 0
            data_[ti] = newspec

    if mw_use is not None:
        pads = np.full((N, margin), False)
        pads[is_rfi_num, :] = True

        pad = np.full(margin, False)
        if side == 'both':
            is_rfi = np.hstack((pads, is_rfi, pads))
            mw_use = np.hstack((pad, mw_use, pad))
        elif side == 'left':
            is_rfi = np.hstack((pads, is_rfi))
            mw_use = np.hstack((pad, mw_use))
        elif side == 'right':
            is_rfi = np.hstack((is_rfi, pads))
            mw_use = np.hstack((mw_use, pad))

    return data_, freq_, is_rfi, mw_use, margin

# Cell
def get_margin_num(margin, side):
    if side == 'both':
        margin1 = margin
        margin2 = -margin
    elif side == 'left':
        margin1 = margin
        margin2 = 0
    elif side == 'right':
        margin2 = margin
        margin1 = 0
    return margin1, margin2


def replace_rfi_lower(data, freq, method, times_lower_thr=3, times_lower=None,
                      rms_sigma=5, rms_frange=None):

    from .markRFI import real_rms
    RMS = real_rms(data[0, :], freq, sigma=rms_sigma, rms_vrange=rms_frange)

    thr_lower = RMS*times_lower_thr
    exceed_use = (np.abs(data) > thr_lower)
    from ..utils.misc import extend_Trues
    exceed_use = extend_Trues(exceed_use, ext_add=10, leng_lim=20, axis=-1)

    data_low = deepcopy(data)
    if method == 'lower':
        data_low[exceed_use] = data[exceed_use]/times_lower
    elif method == 'set_zeros':
        data_low[exceed_use] = 0
    elif method == 'set_noise':
        data_low[exceed_use] = np.random.normal(scale=RMS, size=data.shape)[exceed_use]
    return data_low


def replace_rfi(data, freq, time_rfi=None, method='near_ripple',
                mw_frange=None, **rep_args):

    log.info(f"Replace RFI with {method} method ...")
    if mw_frange is None:
        if (max(freq) <= 1419) | (min(freq) >= 1422):
            print("don't contain MW")
            mw_use = np.zeros_like(freq, dtype='bool')
        else:
            mw_use = None
    else:
        mw_use = (freq >= mw_frange[0]) & (freq <= mw_frange[1])
    is_excluded = None
    if method == 'near_ripple':
        data_rep, is_excluded = repalce_near(data, freq, time_rfi=time_rfi, mw_use=mw_use, **rep_args)
    elif (method == 'lower') or (method == 'set_zeros') or (method == 'set_noise'):
        data_rep = replace_rfi_lower(data, freq, method=method, **rep_args)
    else:
        raise ValueError("Unsupport replace RFI method!")

    return data_rep, is_excluded

# Cell
def find_loc(amp, x, amp_thr, xlim, rip_mhz=True, Print=True):
    """
    amp: fft amptitude
    x: fft xfreq
    amp_thr: amp threshold
    rip_mhz: bool
    xlim: find max between xlim[0] and xlim[1]
    """
    loc = None
    sw_mhz = None
    if rip_mhz:
        u = (amp > amp_thr) & (x > xlim[0]) & (x < xlim[1])

        if np.sum(u) > 0:
            amp_ = np.zeros_like(x)
            amp_[u] = amp[u]

            loc = np.argmax(amp_)
            sw_mhz = 1/x[loc]
            if Print:
                print(f'find ripple {sw_mhz:.5f} MHz,{1/sw_mhz:.5f} mu s,locate in {loc}')
        else:
            rip_mhz = False

    return loc, sw_mhz, rip_mhz

# Cell
class SW_FFT(object):
    def __init__(self, s2p, freq, workers=10):
        if s2p.ndim != 2:
            raise(ValueError('input s2p should be 2d'))
        if len(freq) != s2p.shape[1]:
            raise(ValueError('len(freq) need equals to s2p.shape[1]'))
        self.s2p = s2p
        self.freq = freq
        self.workers = workers

    def do_fft(self,):
        A_data = fft.rfft(self.s2p, n=len(self.freq), axis=1, workers=self.workers)
        #self.complex_num = A_data
        self.x = fft.rfftfreq(self.s2p.shape[1], self.freq[1]-self.freq[0])
        self.period = 1/self.x
        self.amp = np.abs(A_data)
        self.phi = np.angle(A_data,)

    @property
    def normal_amp(self):
        N = len(self.freq)
        data = deepcopy(self.amp) # K
        data[:,0] = data[:,0] / N
        data[:,1:] = data[:,1:] / (N / 2)
        return data * 1e3 # mK

    def gen_amp_thr_s(self, amp_thr_mean_factor=1.05, amp_thr_solo_factor=1.4,
                      is_excluded_mean=None):

        if is_excluded_mean is None:
            is_excluded_mean = np.full(len(self.amp), False, dtype=bool)
        self.amp_mean_t = np.nanmean(self.amp[~is_excluded_mean], axis=0)
        self.amp_thr_mean = np.nanmedian(self.amp_mean_t) * amp_thr_mean_factor

        self.amp_thr_solo = np.nanmedian(self.amp_mean_t) * amp_thr_solo_factor

    def find_sw_loc(self, xlims=[[.90, .95], [1.8, 1.9]]):
        """
        find whether a standing wave exists in the xlim range
        can run many times
        """
        xlims = np.array(xlims)
        if xlims.ndim == 1:
            xlims = xlims[None, :]
        self.bingo_list = []
        self.loc_list = []
        self.sw_mhz_list = []
        for xlim in xlims:
            loc, sw_mhz, bingo = find_loc(self.amp_mean_t, self.x, self.amp_thr_mean, rip_mhz=True, xlim=xlim)
            self.loc_list += [loc, ]
            self.sw_mhz_list += [sw_mhz, ]
            self.bingo_list += [bingo, ]

    def find_sw_chans(self, nchans,):
        """
        find channels between loc-nchan and loc+nchan
        run after self.find_sw_loc, can run many times
        """
        if np.isscalar(nchans):
            nchans = np.full(len(self.bingo_list), nchans)
        elif len(nchans) == 1:
            nchans = np.full(len(self.bingo_list), nchans[0])
        elif len(nchans) != len(self.bingo_list):
            raise(ValueError('the number nchans is not same with xlim'))
        if not hasattr(self, 'is_use_sw_chan'):
            self.is_use_sw_chan = np.full(len(self.x), False)
        for i, bingo in enumerate(self.bingo_list):
            if bingo:
                self.is_use_sw_chan |= (np.abs(np.arange(len(self.x))-self.loc_list[i]) < nchans[i])

    def choose_and_ifft(self, method='interpolate', inplace_amp=True, sw_base=True):
        """
        inplace_amp: make a copy of self.amp
        """
        amp = self.amp if inplace_amp else np.copy(self.amp)
        if not hasattr(self, 'is_use_sw_chan'):
            self.is_use_sw_chan = np.full(len(self.x), False)
        is_use_sw_chan_amp = self.is_use_sw_chan & (amp > self.amp_thr_solo)

        if method == 'interpolate':
            # interpolate
            from scipy.interpolate import interp1d
            for ti in tqdm(range(amp.shape[0])):
                x_mask = self.x[~is_use_sw_chan_amp[ti]]
                amp_solo_mask = amp[ti][~is_use_sw_chan_amp[ti]]
                amp_solo_interp = interp1d(x_mask, amp_solo_mask, kind='linear')  # ,fill_value="extrapolate")
                amp[ti][is_use_sw_chan_amp[ti]] -= amp_solo_interp(self.x[is_use_sw_chan_amp[ti]])
        elif method == 'all':
            # no need to process amp[is_use_sw_chan_amp]
            pass
        if sw_base:
            amp[:, 1:][~is_use_sw_chan_amp[:, 1:]] = 0
        else:
            amp[~is_use_sw_chan_amp] = 0
        self.sw = np.real(fft.irfft(amp*np.exp(1j*self.phi), n=len(self.freq), axis=1, workers=self.workers))
        #self.is_use_sw_chan_amp = is_use_sw_chan_amp

# Cell
def get_sw_conf(chan_wide, chan_narr):
    sw_conf = {
        '1mhz': {'xlims': [[.90, .95], [1.8, 1.9]], 'nchans': [chan_wide, chan_narr]},
        '2mhz': {'xlims': [[.5, .6], ], 'nchans': [chan_wide]},
        '0_04mhz': {'xlims': [[25, 26]], 'nchans': [chan_narr]},
    }
    return sw_conf


def fit_sw_fft(s1p, freq, nproc, amp_thr_mean_factor=1.05, amp_thr_solo_factor=1.4, is_excluded_mean=None,
               chan_wide=5, chan_narr=3, sw_periods=['1mhz', '2mhz', '0_04mhz'],
               sw_base=True, choose_method='all'):
    """
    amp_thr_mean_factor: above mean amptitude threshold will be chosed
    amp_thr_solo_factor: above solo amptitude threshold will be chosed
    is_excluded_mean: bool, exclude large RFI
    chan_wide: channel numbers near 1mhz to be chosed (wide)
    chan_narr: channel numbers near 1mhz to be chosed (narrow)
    sw_periods: remove ripple (1mhz: 1.08mhz, 2mhz:1.92mhz, 0_04mhz: 0.039 mhz)
    choose_method: use 'all' modes or 'interpolate' from nearby modes
    sw_base: if True, remove constant components / the base frequency (0 \mu s)
    """

    sw_conf = get_sw_conf(chan_wide, chan_narr)
    sw = SW_FFT(s1p, freq, nproc)
    sw.do_fft()
    sw.gen_amp_thr_s(amp_thr_mean_factor=amp_thr_mean_factor, amp_thr_solo_factor=amp_thr_solo_factor,
                     is_excluded_mean=is_excluded_mean)
    if 'none' not in sw_periods:
        for key in sw_periods:
            sw.find_sw_loc(xlims=sw_conf[key]['xlims'])
            sw.find_sw_chans(nchans=sw_conf[key]['nchans'])
    sw.choose_and_ifft(method=choose_method, inplace_amp=True, sw_base=sw_base)
    return sw.sw






__all__ = ['FastRawData', 'adjust_frange', 'get_str_aft_nB', 'HiFASTData', 'FastRawSpec', 'CalOnOff', 'plot_sep',
           'mean_a', 'CalOnOffA', 'PositionSwitch']


"""
@author: yixiancao
@author: YingjieJing
"""
import os
import re
from glob import glob

import numpy as np
import scipy.interpolate as interp
import h5py
from astropy.io import fits

import __main__
if hasattr(__main__, '__file__'):
    import matplotlib
    matplotlib.use('Agg')
from matplotlib import pyplot as plt


from ..utils.tcal import read_tcal
from ..utils.misc import smooth_axis1_d3, down_sample, median_filter_axis1_d3
from ..utils.misc import smooth1d
from ..utils.io import save_specs_hdf5, MjdChanPolar_to_PolarMjdChan
from ..utils.io import H5FitsRead


class FastRawData(object):
    """ RAW data from FAST observations.

    Attributes
    ----------
    filename: str
        Name of Raw data file (FITS file)
    fileinfo: dict
        Infomation inferred from the filename
    obs: dict
        Basic observation informaiton from observation log (NOW hd0 header).

    Methods
    ----------
    listInfo():
        Display the basic infomation about the data file.

    """


    def __init__(self, filenames, frange=None, dfactor=None, med_filter_size=None, verbose=False):
        """
        Parameters
        ----------
        filenames : list
            FAST fits file path name
        frange : [min max], optional
            Frequency range
        dfactor : str 'W' or int or None, optional
            if 'W', downsampling 'N' or 'F' to 'W' band; if int, downsampling the spectra with a factor of dfactor
        med_filter_size : int, optional
            median filter spectra with size of this value
        verbose : bool, optional
            Default is False
        """
        self.filenames = filenames
        self.verbose = verbose
        if self.verbose:
            print('files:')
            for filename in self.filenames:
                print(filename)
        if self.filenames[0].endswith('.fits'):
            self.hduls = [fits.open(filename) for filename in self.filenames]
            self.ftype = 'fits'
        elif self.filenames[0].endswith('.hdf5'):
            self.hduls = [H5FitsRead(filename) for filename in self.filenames]
            self.ftype = 'hdf5'
        else:
            raise(ValueError('File not support'))

        self.hd0s = [hdul[0].header for hdul in self.hduls]
        self.hd1s = [hdul[1].header for hdul in self.hduls]
        self.lens= [header['NAXIS2'] for header in self.hd1s]
        self.inds = np.arange(np.sum(self.lens))

        if isinstance(dfactor, str):
            if dfactor.upper() == 'W':
                dfactor = 16
            elif dfactor.isdigit():
                dfactor = int(dfactor)
            else:
                raise(ValueError("dfactor should be 'W' or an int"))
        # adjust frange
        if dfactor == 16 and frange is not None:
            frange = adjust_frange(frange)

        self.frange = frange
        self.dfactor = dfactor
        self.med_filter_size = med_filter_size
        self._get_freq()
        if frange is not None:
            is_use = (self.freq<= frange[1] ) & (self.freq>= frange[0])
            self.freq_use = self.freq[is_use]
        else:
            self.freq_use= self.freq
        if len(self.freq_use) == 0:
            raise(ValueError(f"frange {frange} is not in freq range of {self.freq}"))
        if self.dfactor is not None:
            self.freq_use = down_sample(self.freq_use[None,:,None], self.dfactor)[0,:,0]

    def ls_fields(self,):
        """
        list fields in fits hdu1.data
        """
        print(self.hduls[0][1].data.dtype)

    def get_mjds(self):
        mjds = np.hstack([hdul[1].data['UTOBS'] for hdul in self.hduls])
        return mjds

    def _get_freq(self, center_corr=None):
        tdata = [self.hduls[0][1].data]
        nchan = tdata[0]['NCHAN'][0]
        freq0 = tdata[0]['FREQ'][0]
        chanwidth = tdata[0]['CHAN_BW'][0]
        freq=(np.arange(nchan))*chanwidth
        freq=freq+freq0
        if center_corr is not None:
            freq += center_corr # using center frequency
        self.freq= freq

    def get_field(self, _inds, field='DATA', close_file=False):
        """"
        get field with index
        """
        # if input _inds is empty, return empty
        if 0 in _inds.shape:
            if field == 'DATA':
                return np.empty(_inds.shape + (len(self.freq_use),2))
        # flat _inds
        _input_ndim = _inds.ndim
        if _input_ndim > 1:
            _in_shape= _inds.shape
            _inds= _inds.flatten()
        # h5py needs flattened _inds is monotone increasing
        if self.ftype == 'hdf5':
            if (np.diff(_inds) <= 0).any():
                # load all data to the numpy array first
                load_all_first = True
            else:
                load_all_first = False
        else:
            load_all_first = False
        # determine the file and index of the spec in
        lens_cum = np.hstack([0,np.cumsum(self.lens)])
        ifile =  np.searchsorted(lens_cum, _inds + 1, side='left') - 1
        ind_ifile = _inds- lens_cum[ifile]
        ifile_uni, ifile_num = np.unique(ifile, return_counts=True)
        ind_ifile_list= np.split(ind_ifile, np.cumsum(ifile_num)[:-1])
        # load data
        j_list=[]
        for j in ifile_uni:
            try:
                # test if file is opened
                if self.ftype == 'fits':
                    self.hduls[j][1].data.shape
                elif self.ftype == 'hdf5':
                    self.hduls[j][1].data[field].shape
            except:
                if self.ftype == 'fits':
                    self.hduls[j] = fits.open(self.filenames[j], memmap=True, lazy_load_hdus=True)
                elif self.ftype == 'hdf5':
                    self.hduls[j] = H5FitsRead(self.filenames[j])
                j_list += [j,]
        if field == 'DATA':
            if self.frange is not None:
                is_use = (self.freq <= self.frange[1] ) & (self.freq >= self.frange[0]) # can't use self.freq_use
                ind_use= np.where(is_use)[0]
                # freq axis (ind_use) need use "slice" to index, coz ii is already a array; ind_use is continuous
                if not load_all_first:
                    data = np.vstack([self.hduls[i][1].data[field][ii, ind_use[0]:ind_use[-1]+1, :2] for i, ii in zip(ifile_uni,ind_ifile_list)])
                else:
                    data = np.vstack([self.hduls[i][1].data[field][:][ii, ind_use[0]:ind_use[-1]+1, :2] for i, ii in zip(ifile_uni,ind_ifile_list)])
            else:
                if not load_all_first:
                    data = np.vstack([self.hduls[i][1].data[field][ii, :, :2] for i, ii in zip(ifile_uni,ind_ifile_list)])
                else:
                    data = np.vstack([self.hduls[i][1].data[field][:][ii, :, :2] for i, ii in zip(ifile_uni,ind_ifile_list)])
            #
            if self.med_filter_size is not None:
                data = median_filter_axis1_d3(data, self.med_filter_size)
            # Downsampling
            if self.dfactor is not None:
                data = down_sample(data.astype('float64'), self.dfactor)
        else:
            data = np.hstack([self.hduls[i][1].data[field][ii] for i, ii in zip(ifile_uni,ind_ifile_list)])
        #check if the length in the header is correct
        for i_tmp in ifile_uni:
            if self.ftype == 'fits':
                  leng = self.hduls[i_tmp][1].data.shape[0]
            elif self.ftype == 'hdf5':
                  leng = self.hduls[i_tmp][1].data[field].shape[0]
            if not self.lens[i_tmp] == leng:
                raise(ValueError(f"length in the header is not equal to the data shape:", self.filenames[i_tmp]))
        [self.hduls[j].close() for j in j_list]
        if close_file:
            [self.hduls[i].close() for i in ifile_uni]

        if _input_ndim > 1:
            return data.reshape(_in_shape + data.shape[1:])
        else:
            return data

    @property
    def fileinfo(self):
        fitsname = dict(value = self.filenames[0].split('/')[-1], desc = "First FITS file name")
        dummy = fitsname['value']
        tinfo = dummy.split('-')[0]
        tinfo = tinfo.split('_')
        #project = dict(value = 'FAST_M31', desc = "Project name")
        source = dict(value = tinfo[0], desc = "Source name")
        ipos = dict(value  = int(tinfo[1][1:]), desc = "Position number")
        #iobs = dict(value = None, desc = "Observation number")
        track_mode = dict(value = tinfo[2], desc = "Observation mode") # e.g. Tracking, drifting

        finfo = dummy.split('-')[1]
        finfo = finfo.split('_')
        ibeam = dict(value = int(finfo[0][1:]), desc = "Beam number")
        #ifile = dict(value = int(finfo[-1][0:4]), desc = "File number") #ith file for the observation

        filetype = dict(value = 'psr' if len(finfo) < 3 else 'spec' +  finfo[1],
                         desc = "Observation type")

#         fileinfo = dict(fitsname = fitsname, project=project, source = source, ipos = ipos, track_mode = track_mode,
#                         iobs =iobs, ibeam = ibeam, ifile = ifile, filetype = filetype)
        fileinfo = dict(fitsname = fitsname, source = source, ipos = ipos, track_mode = track_mode,
                         ibeam = ibeam, filetype = filetype)

        return fileinfo

    def listInfo(self, list_type = 'all'):
        finfo = self.fileinfo
        print("FAST RAW DATA FILE Summary")
        print("==================")
        for key, value in finfo.items():
            print (value['desc'], ':' , value['value'])

    #@property
    def obs(self,chunk_num=0):
        """ Dictionary of observation parameters.
            From HD0 or from observation log
        """
        hd0 = self.hd0s[chunk_num]
        time = {"t_obs": dict(value  = hd0['DATE'],
                              desc = 'Observation time (starting to record data)') ,
                "t_samp": dict(value = None,
                               desc = 'Sampling time interval (s)')
                }
        track = {"Mode": dict(value = self.fileinfo['track_mode'],
                                    desc =  "Tracking mode (e.g. tracking, drifting, etc)"),
                "RA": dict(value = np.double(0.0), desc = "Tracking RA"),
                "DEC": dict(value = np.double(0.0), desc = "Tracking DEC")}

        noise_diode = {"Mode": dict(value = "",  desc = "Noise diode mode (Modulate, ON, or OFF)"),
                      "Power": dict(value = "",  desc = "Noise diode power (High/Low)"),
                      "Delay": dict(value = 0, desc =  "Noise Delay in unit of tunit"),
                      "On": dict(value = 0,  desc = "Noise diode ON in unit of tunit"),
                      "Off":  dict(value = 0,  desc = "Noise diode OFF in unit of tunit"),
                      "tunit":  dict(value = 4e-9, desc =  "Units of delay/on/off time (s)")
                      } #
        receiver = {"Frontend": dict(value = '', desc =  "Receiver frontend (e.g. 19BEAM)"),
                    "rfgain": dict(value = 0.0, desc =  "Receiver rfgain (dB)"),
                    "dgain": dict(value = 0.0, desc = "Receiver dgain"),
                    "Backend": dict(value = '', desc = "Receiver Backend ID (e.g. 'MB4K')")
                    }

        obs = dict(time = time,
               track = track,
               noise_diode = noise_diode,
               receiver = receiver)

        return obs


def adjust_frange(frange):
    """
    adjust freq range frange to let the F band downsample freq is same with W band
    """

#     # W band
#     nchan = 65536
#     chanwidth = 0.00762939453125
#     freq0 = 1000.0035762786865
#     freq_W = np.arange(nchan)*chanwidth + freq0 + 0.000476837158203125/2.

    # F band
    nchan = np.array(1048576, dtype='int64')
    chanwidth = np.array(0.000476837158203125, dtype='float64')
    freq0 = np.array(1000.0, dtype='float64')

    chanwidth_F = np.array(0.000476837158203125, dtype='float64')
    freq_F = np.arange(nchan)*chanwidth + freq0 + chanwidth_F/2.

    offset = chanwidth_F/4.
    ind_close = np.argmin(abs(freq_F - np.array(frange)[:,None]), axis=1)
    ind_close = ((ind_close+1)//16+1)*16-1
    frange_new = (freq_F[ind_close[0]] + offset, frange[1])
#     width = int(np.round(abs(np.log10(offset)))+1)
#     print(ind_close)
#     print(freq_F)
#     print(f"%.{width}f %.{width}f" % frange_new)
    return frange_new


def get_str_aft_nB(path, nth=1):
    """
    path
    nth: beginning with 1
    """
    basename = os.path.basename(path)
    ind = re.search(r'[0-9][0-1]M-', basename[::-1]).start()
    str_list = basename[-ind:].split('-')[1:]
    if len(str_list) >= nth:
        return str_list[nth-1]
    else:
        return None


class HiFASTData(FastRawData):
    """ data in HiFAST.

    """


    def __init__(self, filenames, frange=None, dfactor=None, med_filter_size=None, verbose=False):
        """
        Parameters
        ----------
        filenames : list
            FAST fits file path name
        frange : [min max], optional
            Frequency range
        dfactor : str 'W' or int or None, optional
            if 'W', downsampling 'N' or 'F' to 'W' band; if int, downsampling the spectra with a factor of dfactor
        med_filter_size : int, optional
            median filter spectra with size of this value
        verbose : bool, optional
            Default is False
        """
        self.filenames = filenames
        self.verbose = verbose
        if self.verbose:
            print('files:')
            for filename in self.filenames:
                print(filename)

        self.ftype = 'hdf5' # must be 'hdf5'
        self.hduls = [H5FitsRead(filename, kT='Power', hdu_names=['S'], DATA_key='Power') for filename in self.filenames]
        for hdul in self.hduls:
            hdul.hdus[1] = hdul.hdus['S']
            hdul.hdus[1].name = '1'

        self.hd0s = [{} for hdul in self.hduls]
        self.hd1s = [hdul[1].header for hdul in self.hduls]
        self.lens= [hdul[1].data['mjd'].shape[0] for hdul in self.hduls]
        self.inds = np.arange(np.sum(self.lens))

        if isinstance(dfactor, str):
            if dfactor.upper() == 'W':
                dfactor = 16
            elif dfactor.isdigit():
                dfactor = int(dfactor)
            else:
                raise(ValueError("dfactor should be 'W' or an int"))
        # adjust frange
        if dfactor == 16 and frange is not None:
            frange = adjust_frange(frange)

        self.frange = frange
        self.dfactor = dfactor
        self.med_filter_size = med_filter_size
        self._get_freq()
        if frange is not None:
            is_use = (self.freq<= frange[1] ) & (self.freq>= frange[0])
            self.freq_use = self.freq[is_use]
        else:
            self.freq_use= self.freq
        if len(self.freq_use) == 0:
            raise(ValueError(f"frange {frange} is not in freq range of {self.freq}"))
        if self.dfactor is not None:
            self.freq_use = down_sample(self.freq_use[None,:,None], self.dfactor)[0,:,0]

    def ls_fields(self,):
        """
        list fields in fits hdu1.data
        """
        print(self.hduls[0][1].data.keys())

    def get_mjds(self):
        mjds = np.hstack([hdul[1].data['mjd'] for hdul in self.hduls])
        return mjds

    def _get_freq(self, center_corr=None):
        # center_corr already in freq
        self.freq = self.get_field(field='freq')

    def get_field(self, _inds=slice(None), field='DATA', close_file=False):
        """
        """
        # if input _inds is empty, return empty
        if (not isinstance(_inds, slice)) and 0 in _inds.shape:
            if field == 'DATA':
                # if input _inds is empty, return empty
                return np.empty(_inds.shape + (len(self.freq_use),2))
        # field used get_field to read
        if field in ['DATA', 'ra', 'dec', 'mjd', 'UTOBS']:
            if field == 'UTOBS':
                field = 'mjd'
            res = super().get_field(_inds, field=field, close_file=close_file)
        elif field in ['freq']:
            # no need _inds
            res = np.hstack([hdul[1].data[field][:] for hdul in self.hduls])
        else:
            raise(NotImplementedError)
        return res

    def fileinfo(self):
        raise(NotImplementedError)

    def listInfo(self, list_type = 'all'):
        raise(NotImplementedError)

    def obs(self,chunk_num=0):
        raise(NotImplementedError)


class FastRawSpec(object):
    def __new__(cls, *args, **kwargs):
        if 'fname_part' in kwargs.keys():
            fname_part = kwargs['fname_part']
        else:
            fname_part = args[0]
        if fname_part.endswith('.hdf5') and str(get_str_aft_nB(fname_part, 2)).startswith('specs'):
            cls_base = HiFASTData
        else:
            cls_base = FastRawData
        class DynamicC(cls, cls_base):
            pass
        return super().__new__(DynamicC)

    def __init__(self, fname_part, start=1, stop=None,
                 frange=None,
                 dfactor=None, med_filter_size=None,
                 verbose=False,
                 smooth='gaussian', s_para={'s_sigma':5},
                 noise_mode=None, noise_date='auto'):
        """
        fname_part: str; one of chunk file name; '*.hdf5' or '*.fits'
        start: int; chunk file start
        stop: int; chunk file stop
        frange: frequency range
        dfactor: 'W' or int; if 'W', downsampling 'N' or 'F' to 'W' band; if int, downsampling the spectra with a factor of dfactor
        """
        # generate fits filename list
        if fname_part.endswith('.fits'):
            ftype = 'fits'
        elif fname_part.endswith('.hdf5'):
            ftype = 'hdf5'
        else:
            raise(ValueError('File not support'))
        if ftype == 'hdf5' and str(get_str_aft_nB(fname_part, 2)).startswith('specs'):
            filenames = [fname_part,]
            fname_part = fname_part.strip('.hdf5')
        else:
            fname_part = re.sub('[0-9]{4}\.%s\Z' % ftype, '', fname_part)
            if stop is not None:
                if stop < 1:
                    raise(ValueError('input stop < 1'))
            else:
                stop = len(glob(fname_part + f'[0-9][0-9][0-9][0-9].{ftype}'))
            filenames = [f"{fname_part}{i:04d}.{ftype}" for i in range(start,stop+1)]
        if len(filenames) == 0:
            raise(OSError(f"can not find file, please check fname_part:{fname_part}"))
        self.fname_part = fname_part
        super().__init__(filenames, frange=frange, dfactor=dfactor, med_filter_size=med_filter_size, verbose=verbose)

        self.nB= int(re.findall(r'-M[0-1][0-9]', fname_part)[-1][2:])
        self.smooth = smooth
        self.s_para = s_para
        self.noise_mode = noise_mode
        self.noise_date = noise_date

    def _get_freq(self,):
        # only for W,F,N band
        #freq += 0.000476837158203125/2 # using center frequency
        super()._get_freq(center_corr = 0.000476837158203125/2)

    def _get_smoothed(self, power, freq=None, use_ndimage=False, check_nan=False):
        """
        power: ndim 3
        freq: if is None, use self.freq_use
        """
        smooth = self.smooth
        if freq is None:
            freq = self.freq_use
        s_para = self.s_para
        power = power.astype('float64')
        if smooth=='gaussian':
            sigma = s_para['s_sigma']/(np.nanmax(freq) - np.nanmin(freq))*len(freq)
            if use_ndimage:
                from scipy import ndimage
                s_power = ndimage.gaussian_filter1d(power, sigma=sigma, axis=1)
            elif check_nan:
                is_ = np.isfinite(power)
                power[~is_] = 0
                s_power = smooth_axis1_d3(power, method=smooth, sigma=sigma)
                s_power /= smooth_axis1_d3(is_.astype('float64'), method=smooth, sigma=sigma)
                s_power[np.isinf(s_power)] = np.nan
                s_power[~is_] = np.nan
                power[~is_] = np.nan
            else:
                s_power = smooth_axis1_d3(power, method=smooth, sigma=sigma)

        elif smooth=='poly':
            s_power = smooth_axis1_d3(power, method=smooth, x=freq, deg= s_para['s_deg'])
        else:
            raise(ValueError('smooth method not support'))
        return s_power

    def get_Tcal_s(self,):
        """
        load T cal from noise file, smoothed
        """
        if self.noise_date == 'auto':
            _mjd = self.get_field(np.array([0,]), field='UTOBS')[0]
        else:
            _mjd = None
        tc_freq, tc_, self.tcal_file = read_tcal(self.nB, mode= self.noise_mode, date=self.noise_date, mjd=_mjd)
        if self.frange is not None:
            is_use = (tc_freq <= self.frange[1] ) & (tc_freq >= self.frange[0])
        else:
            is_use = np.full(len(tc_freq), True, dtype=bool)
        tc_= tc_[:, is_use]
        tc_freq= tc_freq[is_use]
        tc= self._get_smoothed(tc_.T[None,:,:], freq=tc_freq)
        tc_inter = interp.interp1d(tc_freq, tc, axis=1, kind='linear', fill_value ='extrapolate')(self.freq_use)
        return tc_inter

    def gen_out_name_base(self, outdir):
        fname_add = os.path.basename(os.path.dirname(os.path.abspath(self.fname_part)))
        self.out_name_base = os.path.join(outdir, f"{os.path.basename(self.fname_part)[:-1]}-{fname_add}")


class CalOnOff(FastRawSpec):
    """
    """
    def __init__(self, fname_part, n_delay, n_on, n_off, med_filter_size_cal=5, p_cal_fname=None, **kwargs):
        """

        Parameters
        ----------
        fname_part : str
            file name with out chunk postfix, for example:
            '/data/inspur_disk06/fast_data/3047/M31_Halo_Drift/20200106/M31_Halo_Drift_1_arcdrift-M02_F_'
        start : int
            first chunk number; Tcal on should be in the beginning of this file.
        stop : int
            last chunk number
        n_on : int
            Tcal_on Time divided by Sampling Time
        n_off : int
            Tcal_off Time divided by Sampling Time
        frange : list; [min,max]
            range of freq to use

        """
        super().__init__(fname_part, **kwargs)
        self.inds = np.arange(np.sum(self.lens))
        self.n_on, self.n_off= n_on, n_off
        self.n_delay = n_delay
        if n_delay >= 1 and self.n_on >=2:
            self.tcal_offset = True
        else:
            self.tcal_offset = False
        self.med_filter_size_cal = med_filter_size_cal
        if p_cal_fname is not None:
            self.p_cal_f = h5py.File(p_cal_fname, 'r')
        else:
            self.p_cal_f = None
        self.sep_on_off_inds()
        # set init plot as False
        # self.plot = False

    def get_extra(self,):
        """
        get extra info
        """
        extra = {}

        close_cal= np.full(self.inds.shape, False, dtype=bool)
        ind_tmp= self.inds_ton[:,0]-1
        if ind_tmp[0] ==-1:
            close_cal[ind_tmp[1:]] = True
        else:
            close_cal[ind_tmp] = True
        ind_tmp= self.inds_ton[:,-1]+1
        if ind_tmp[-1] > len(close_cal)-1:
            close_cal[ind_tmp[:-1]] = True
        else:
            close_cal[ind_tmp] = True
        extra['next_to_cal'] = close_cal

        is_on = np.full(self.inds.shape, False, dtype=bool)
        is_on[self.inds_on] = True
        extra['is_on'] = is_on

        is_delay = np.full(self.inds.shape, False, dtype=bool)
        is_delay[self.inds_delay] = True
        extra['is_delay'] = is_delay

        return extra

    def sep_on_off_inds(self):
        """
        separate on and off spec inds
        """
        n_delay, n_on, n_off = self.n_delay, self.n_on, self.n_off
        inds = self.inds
        nth = (inds - n_delay) % (n_on+n_off)
        self.inds_on = self.inds[(nth<n_on) & (inds>=n_delay)]
        self.inds_off = self.inds[(nth>=n_on) | (inds<n_delay)] # delay as off
        self.inds_delay = self.inds[inds<n_delay]
        #
        self._get_tcal_inds()

    def _get_tcal_inds(self):
        """
        find index of the samples used to calculate power of Noise diode
        """
        # drop incomplete tcal on
        n_remainder = len(self.inds_on)%(self.n_on)
        # Tcal_on index
        inds_ton= self.inds_on[:(len(self.inds_on)-n_remainder)].reshape((-1,self.n_on))
        # Tcal_off index
        nbef = self.n_on//2
        inds_toff_bef = inds_ton[:,:nbef] - nbef
        inds_toff_aft = inds_ton[:,nbef:] + self.n_on - nbef
        # fix first
        is_exceed = inds_toff_bef[0] < 0
        n_exceed = np.sum(is_exceed)
        if 0 < n_exceed < len(is_exceed):
            inds_toff_bef[0][is_exceed] = np.full(np.sum(is_exceed), inds_toff_bef[0][~is_exceed][0])
        if 0 < n_exceed and n_exceed == len(is_exceed):
            if self.tcal_offset:
                # drop first
                inds_ton = inds_ton[1:]
                inds_toff_bef = inds_toff_bef[1:]
                inds_toff_aft = inds_toff_aft[1:]
            else:
                inds_toff_bef[0][is_exceed] = inds_ton[0][-1] + 1
        # fix last
        is_exceed = inds_toff_aft[-1] > self.inds[-1]
        n_exceed = np.sum(is_exceed)
        if 0 < n_exceed < len(is_exceed):
            inds_toff_aft[-1][is_exceed] = np.full(np.sum(is_exceed), inds_toff_aft[-1][~is_exceed][-1])
        if  0 < n_exceed and n_exceed == len(is_exceed):
            if self.tcal_offset:
                # drop last
                inds_ton = inds_ton[:-1]
                inds_toff_bef = inds_toff_bef[:-1]
                inds_toff_aft = inds_toff_aft[:-1]
            else:
                inds_toff_aft[-1][is_exceed] = inds_ton[-1][0] - 1

        self.inds_ton = inds_ton
        self.inds_toff_bef = inds_toff_bef
        self.inds_toff_aft = inds_toff_aft

    def _smooth_cal(self, cal):
        if self.med_filter_size_cal is not None and self.med_filter_size_cal !=0:
            cal = median_filter_axis1_d3(cal, self.med_filter_size_cal)
        cal = self._get_smoothed(cal)
        return cal

    def _get_cal_power_s(self, inds_ton, inds_toff_bef, inds_toff_aft):
        """
        use self._get_cal_power to get the power of cal and smooth it
        """
        p_cal = self._get_cal_power(inds_ton, inds_toff_bef, inds_toff_aft)
        return self._smooth_cal(p_cal)

    def _get_cal_power(self, inds_ton, inds_toff_bef, inds_toff_aft):
        """
        get the power of Noise diode
        """
        p_ton = self.get_field(inds_ton,field='DATA')
        p_toff_aft = self.get_field(inds_toff_aft, field='DATA')
        p_toff_bef = self.get_field(inds_toff_bef, field='DATA')
        try:
            if getattr(self, 'plot', False):
                new_shape= (-1,)+p_ton.shape[-2:]
                figname = self.out_name_base + "-cal.pdf"
                plot_sep(inds_ton.flatten(), np.hstack([inds_toff_bef.reshape(-1),inds_toff_aft.reshape(-1)]),
                         p_ton.reshape(new_shape),
                         np.vstack([p_toff_bef.reshape(new_shape),p_toff_aft.reshape(new_shape)]),
                         figname=figname, n_max=1000, re_tick=True)
        except:
            pass

        if self.tcal_offset and inds_ton.shape[1] <4:
            n = p_ton.shape[-2]
            p_ton = p_ton[range(len(p_ton)), np.argmax(np.mean(p_ton[:,:,n//20:n-n//20,:], axis=(2,3), dtype='float64'), axis=1)]
            p_toff = np.concatenate([p_toff_bef, p_toff_aft],axis=1) # meger p_toff; not np.hstack or np.vstack or np.stack
            n = p_toff.shape[-2]
            p_toff = p_toff[range(len(p_toff)), np.argmin(np.mean(p_toff[:,:,n//20:n-n//20,:], axis=(2,3), dtype='float64'), axis=1)]
        if self.tcal_offset and inds_ton.shape[1] >=4:
            p_ton = np.mean(p_ton[:,1:-1,:,:], axis= 1, dtype='float64')
            p_toff = np.concatenate([p_toff_bef[:,:-1,:,:], p_toff_aft[:,1:,:,:]],axis=1) # meger p_toff
            p_toff = np.mean(p_toff, axis= 1, dtype='float64')
        if not self.tcal_offset:
            p_ton = np.mean(p_ton, axis= 1, dtype='float64')
            p_toff = np.concatenate([p_toff_bef, p_toff_aft],axis=1) # meger p_toff
            p_toff = np.mean(p_toff, axis= 1, dtype='float64')
        p_cal = p_ton.astype('float64') - p_toff.astype('float64')
        return p_cal

    def get_count_tcal(self, inds_on, inds_off):
        """
        get count of tcal
        """
        if self.p_cal_f is not None:
            inds_ton = self.p_cal_f['inds_ton'][:]
        else:
            inds_ton, inds_toff_bef, inds_toff_aft = self.inds_ton, self.inds_toff_bef, self.inds_toff_aft
        # determine which cal power sample to calibrate
        tcal_c= inds_ton[:, inds_ton.shape[1]//2]
        inds_in_tcal_on= np.argmin(abs(inds_on[:,None]- tcal_c[None,:]), axis=1)
        inds_in_tcal_off= np.argmin(abs(inds_off[:,None]- tcal_c[None,:]), axis=1)
        uni= np.unique(np.hstack([inds_in_tcal_on,inds_in_tcal_off]))
        # load cal power
        if self.p_cal_f is not None:
            p_cal_s = self.p_cal_f['p_cal_s'][uni]
        else:
            p_cal_s = self._get_cal_power_s(inds_ton[uni], inds_toff_bef[uni], inds_toff_aft[uni])
        # load on and off power
        p_on = self.get_field(inds_on, 'DATA',)
        p_off = self.get_field(inds_off, 'DATA', close_file=True)
        #print(p_on.dtype, p_off.dtype, p_cal_s.dtype)
        try:
            if getattr(self, 'plot', False):
                figname = self.out_name_base + "-sep.pdf"
                plot_sep(inds_on, inds_off, p_on, p_off, figname=figname)
        except:
            pass

        count_on = p_on.astype('float64') / p_cal_s[np.where(inds_in_tcal_on[:,None] - uni[None,:] ==0, )[1]] - 1 # have subtracted cal
        count_off = p_off.astype('float64') / p_cal_s[np.where(inds_in_tcal_off[:,None] - uni[None,:] ==0, )[1]]
        return count_on, count_off, p_cal_s, inds_ton[uni]

    def __call__(self, outdir='./', step=None, header=None, sep_save=False, save_pcals=False, cali=True):
        """
        get T, mjd etc, and save in hdf5 file

        Parameters
        ----------
        outdir : str
            output directory
        step : int
            number of chunk files to process each time
        header : dict
            add in the output file
        sep_save : bool
            if True, save file every step.
        """
        #
        outfield = 'Ta' if cali else 'Power'
        self.gen_out_name_base(outdir)
        n_step= self.lens[0]*step if step is not None else len(self.inds)
        inds_range= np.append(np.arange(0, self.inds[-1], n_step), self.inds[-1]+1)
        # inds_splited= np.array_split(self.inds, len(self.inds)//n_step)

        Ts=[]
        mjds = []
        tc_inter = self.get_Tcal_s() if cali else 'none'
        extra = self.get_extra()
        p_cal_s_list = []
        inds_ton_list = []

        self.plot = True
        for i in range(len(inds_range)-1):
            print('part', i)
            b, e = inds_range[i:i+2]
            inds_on = self.inds_on[(self.inds_on >=b) & (self.inds_on < e)]
            inds_off = self.inds_off[((self.inds_off >=b) & (self.inds_off < e))]
            inds = np.hstack([inds_on, inds_off])
            sort = np.argsort(inds)
            mjd = self.get_field(inds[sort], field='UTOBS') # have sorted

            if cali:
                count_tcal_res = self.get_count_tcal(inds_on, inds_off)
            else:
                count_tcal_res = [self.get_field(inds_on, 'DATA',), self.get_field(inds_off, 'DATA', close_file=True)]
            T = np.vstack(count_tcal_res[:2])
            if save_pcals and cali:
                p_cal_s_list += [count_tcal_res[2]]
                inds_ton_list += [count_tcal_res[3]]
            else:
                del count_tcal_res
            if cali: T = T*tc_inter
            T = T[sort] #sort T
            T = T.astype('float32') # finally convert to float32

            # update header
            if i==0:
                try:
                    header = {} if header is None else header.copy()
                    header.update({'tcal_file': self.tcal_file})
                except:
                    pass
            if sep_save:
                res={}
                for key in extra.keys():
                    res[key] = extra[key][inds[sort]]
                res['mjd'] = mjd
                res['freq'] = self.freq_use
                res[outfield] = MjdChanPolar_to_PolarMjdChan(T)
                res['Tcal'] = tc_inter
                #print(res)
                outname= self.out_name_base + f"-specs_T_{i:04d}_{i+1:04d}.hdf5"
                res['Header'] = header
                save_specs_hdf5(outname, res, wcs_data_name=outfield)
                print(f"Saved to {outname}")
                del res
            else:
                Ts += [T,]
                mjds += [mjd]
            self.plot = False #only plot for the first loop
        if not sep_save:
            res={}
            res.update(extra)
            res['mjd'] = np.hstack(mjds)
            res['freq'] = self.freq_use
            res[outfield] = MjdChanPolar_to_PolarMjdChan(np.vstack(Ts))
            res['Tcal'] = tc_inter
            outname= self.out_name_base + f"-specs_T.hdf5"
            res['Header'] = header
            save_specs_hdf5(outname, res, wcs_data_name=outfield)
            print(f"Saved to {outname}")
        if save_pcals and cali:
            p_cal_s_res = {}
            inds_ton = np.vstack(inds_ton_list)
            _, ind_uni = np.unique(inds_ton[:,0],return_index=True)
            p_cal_s_res['inds_ton'] = inds_ton[ind_uni]
            p_cal_s_res['p_cal_s'] = np.vstack(p_cal_s_list)[ind_uni]
            outname = self.out_name_base +'_p_cal_s.hdf5'
            print(f"p_cal in {outname}")
            p_cal_s_res['Header'] = header
            save_specs_hdf5(outname, p_cal_s_res)


def plot_sep(inds_on, inds_off, val_on, val_off, axs=None, figname=None, n_max=1500, re_tick=None, vlines_sep=True):
    if axs is None:
        fig, axs= plt.subplots(3,1,figsize=(15,4*3))
    len_freq= val_off.shape[1]
    wbin= np.min([200,len_freq//4])
    # selecting points in "n_max"
    inds_range= np.sort(np.hstack([inds_on, inds_off]))[:n_max][[0,-1]]
    is_use_on = (inds_on >= inds_range[0]) & (inds_on <= inds_range[1])
    is_use_off = (inds_off >= inds_range[0]) & (inds_off <= inds_range[1])
    x1 = inds_on[is_use_on]
    x2 = inds_off[is_use_off]
    # stacking inds as x
    x = np.hstack([x1,x2])
    ind_sort_x = np.argsort(x)
    x_sort = x[ind_sort_x]
    if re_tick is not None:
        is_1_sort = np.hstack([np.full(len(x1),True),np.full(len(x2),False)])[ind_sort_x]
        x_r = np.arange(len(x))
    for ind_s, ax in zip((len_freq-wbin)//4*np.array([1,2,3]), axs):
        y1 = np.median(val_on[is_use_on, ind_s:ind_s+wbin, 0], axis=1)
        y2 = np.median(val_off[is_use_off, ind_s:ind_s+wbin, 0], axis=1)
        y= np.hstack([y1,y2])
        if re_tick is not None:
            y_sort = y[ind_sort_x]
            ax.scatter(x_r[is_1_sort], y_sort[is_1_sort], color= 'r', label='on', marker='.', s=4)
            ax.scatter(x_r[~is_1_sort], y_sort[~is_1_sort], color= 'b', label='off', marker='.', s=4)
            #ax.plot(x_r, y_sort, 'k', lw=1)
            ax.minorticks_on()
            ax.tick_params(axis='x', which='minor', bottom=False)
            if re_tick == 'full':
                tick = x_r
                ax.set_xticks(tick)
                ax.set_xticklabels(x_sort[x_r], rotation='vertical')
            if vlines_sep:
                lines = x_r
                lines = lines[np.where(np.diff(x_sort) !=1)[0]]
                ax.vlines(lines+0.5, *ax.get_ylim(), linewidth=1, alpha=0.7)

        else:
            ax.scatter(x1, y1, color= 'r',label='on', marker='.', s=4)
            ax.scatter(x2, y2, color= 'b',label='off', marker='.', s=4)
            ax.plot(x[ind_sort_x],y[ind_sort_x],'k',lw=1)
        ax.set_ylabel('xx')
        ax.grid()
    ax.legend(frameon = True)
    ax.set_xlabel('Index')
    if figname is not None:
        fig.savefig(figname, bbox_inches='tight')


def mean_a(arr):
    """
    calculate the mean value of 5th to 95th peercent of the input arr
    """
    b, e = np.percentile(arr, [5,95],)
    return np.mean(arr[(arr <=e ) & (arr >= b)], dtype='float64')


class CalOnOffA(CalOnOff):
    """
    """
    def __init__(self, *args, pcal_vary_frac=0.01, **kwargs):
        self.pcal_vary_frac = pcal_vary_frac
        super().__init__(*args, **kwargs)

    def squeeze_freq(self, arr, axis=-1, method='mean'):
        """
        """
        if method == 'mean':
            return np.mean(arr, axis=axis, dtype='float64')
        elif method == 'mean_a':
            return np.apply_along_axis(mean_a, axis, arr)
        elif method == 'median':
            return np.median(arr.astype('float64'), axis=axis)

    def _check_diff(self, inds):
        """
        check if max diff is not larger than self.pcal_vary_frac
        """
        p = self.get_field(inds)
        ps = self.squeeze_freq(p, axis=2, method='mean_a')
        ps_max, ps_min = np.max(ps, axis=1), np.min(ps, axis=1)
        df = (ps_max - ps_min)/ps_min
        is_use = df < self.pcal_vary_frac
        return is_use[:,0] & is_use[:,1]

    def check_cal(self, inds_ton):
        """
        check cal
        """
        n_c = 6
        assert n_c - n_c//2 <= self.n_off

        n_cbef = n_c//2
        n_caft = n_c - n_cbef
        inds_cbef = inds_ton[:,:1] - np.arange(1,n_cbef+1)[::-1]
        inds_caft = inds_ton[:,-1:] + np.arange(1,n_caft+1)

        # fix first
        is_exceed = inds_cbef[0] < 0
        n_exceed = len(np.where(is_exceed)[0])
        if n_exceed >0:
            inds_cbef[0][is_exceed] = inds_caft[0][-1] + np.arange(1,n_exceed+1)
            inds_cbef[0][np.isin(inds_cbef[0], self.inds_on)] = inds_caft[0][-1]
            inds_cbef[0].sort()
        # fix last
        is_exceed = inds_caft[-1] > self.inds[-1]
        n_exceed = len(np.where(is_exceed)[0])
        if n_exceed >0:
            inds_caft[-1][is_exceed] = inds_cbef[-1][0] - np.arange(1,n_exceed+1)[::-1]
            inds_caft[-1][np.isin(inds_caft[-1], self.inds_on)] = inds_cbef[-1][0]
            inds_caft[-1].sort()
        inds_coff = np.hstack([inds_cbef, inds_caft])
        # been sorted !
        inds_coff[0].sort()
        inds_coff[-1].sort()

        # check off
        is_use = self._check_diff(inds_coff)
        # check on
        if inds_ton.shape[1] >= 2:
            is_use &= self._check_diff(inds_ton)
        if self.verbose:
            aba_inds_ton = inds_ton[~is_use]
            if len(aba_inds_ton) > 0:
                print("abandoned Pcal index:")
                print(*aba_inds_ton[:, 0], sep=',')
        return is_use

    def do_check_cal(self, step=None):
        is_use = self.check_cal(self.inds_ton)
        self.inds_ton = self.inds_ton[is_use]
        self.inds_toff_bef = self.inds_toff_bef[is_use]
        self.inds_toff_aft = self.inds_toff_aft[is_use]

    def get_count_tcal(self, inds_on, inds_off):
        """
        get count of tcal
        """
        if self.p_cal_f is not None:
            inds_ton = self.p_cal_f['inds_ton'][:]
        else:
            inds_ton, inds_toff_bef, inds_toff_aft = self.inds_ton, self.inds_toff_bef, self.inds_toff_aft
        # determine which cal power sample to calibrate
        tcal_c= inds_ton[:, inds_ton.shape[1]//2]
        inds_in_tcal_on= np.argmin(abs(inds_on[:,None]- tcal_c[None,:]), axis=1)
        inds_in_tcal_off= np.argmin(abs(inds_off[:,None]- tcal_c[None,:]), axis=1)
        uni= np.unique(np.hstack([inds_in_tcal_on,inds_in_tcal_off]))
        # check cal
        if self.p_cal_f is None:
            is_use = self.check_cal(inds_ton[uni])
            inds_tmp = np.where(is_use)[0]
            if len(inds_tmp) > 0:
                n_c_aft = len(is_use) - inds_tmp[-1] - 1
            else:
                n_c_aft = len(is_use)
            inds_ton_aft = inds_ton[uni[-1]+1 : uni[-1]+1+n_c_aft]
            # check more cal
#             print(inds_ton_aft)
            if len(inds_ton_aft) > 0:
                is_use_aft = self.check_cal(inds_ton_aft)
                inds_ton_use = np.vstack([inds_ton[uni][is_use], inds_ton[uni[-1]+1 : uni[-1]+1+n_c_aft][is_use_aft]])
                inds_toff_bef_use = np.vstack([inds_toff_bef[uni][is_use], inds_toff_bef[uni[-1]+1 : uni[-1]+1+n_c_aft][is_use_aft]])
                inds_toff_aft_use = np.vstack([inds_toff_aft[uni][is_use], inds_toff_aft[uni[-1]+1 : uni[-1]+1+n_c_aft][is_use_aft]])
            else:
                inds_ton_use = inds_ton[uni][is_use]
                inds_toff_bef_use = inds_toff_bef[uni][is_use]
                inds_toff_aft_use = inds_toff_aft[uni][is_use]
#             print(inds_ton_use)
#             print(inds_toff_bef_use)
#             print(inds_toff_aft_use)
        else:
            inds_ton_use = inds_ton[uni]
        # load cal power
        if self.p_cal_f is not None:
            p_cal_s = self.p_cal_f['p_cal_s'][uni]
        else:
            p_cal_s = self._get_cal_power_s(inds_ton_use, inds_toff_bef_use, inds_toff_aft_use)
        # load on and off power
        p_on = self.get_field(inds_on, 'DATA',)
        p_off = self.get_field(inds_off, 'DATA', close_file=True)
        #print(p_on.dtype, p_off.dtype, p_cal_s.dtype)
        try:
            if self.plot:
                figname = self.out_name_base + "-sep.pdf"
                plot_sep(inds_on, inds_off, p_on, p_off, figname=figname)
        except AttributeError:
            pass

        # redetermine which cal to use
        tcal_c = inds_ton_use[:, inds_ton_use.shape[1]//2]
        inds_in_tcal_on = np.argmin(abs(inds_on[:,None]- tcal_c[None,:]), axis=1)
        inds_in_tcal_off = np.argmin(abs(inds_off[:,None]- tcal_c[None,:]), axis=1)
        count_on = p_on.astype('float64') / p_cal_s[inds_in_tcal_on] - 1 # have subtracted cal
        count_off = p_off.astype('float64') / p_cal_s[inds_in_tcal_off]
        return count_on, count_off, p_cal_s, inds_ton_use


class PositionSwitch(CalOnOff):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # freq: self.freq_use
        self.mjd_a = self.get_mjds()
        # noise-diode on
        self.mjd_on = self.mjd_a[self.inds_on]
        # noise-diode off
        self.mjd_off = self.mjd_a[self.inds_off]

        # power of noise-diode on
        self.p_on = self.get_field(self.inds_on, field='DATA')  # power, shape (mjd,freq,polarization)
        # power of noise-diode off
        self.p_off = self.get_field(self.inds_off, field='DATA')  # power, shape (mjd,freq,polarization)

        self.Tcal_s = self.get_Tcal_s()


    @staticmethod
    def in_range(arr, starts, length):
        """
        arr: shape (m,)
        starts: shape (n,)
        length: scalar
        """
        is_in =  (arr[:,None] > starts[None,:]) & (arr[:,None] < (starts+length)[None,:])
        return np.logical_or.reduce(is_in, axis=1)

    def sep(self, t_src, t_ref, n_repeat, t_change=30):
        """
        second
        """
        #######################
        # on: noise-diode on
        # off: noise-diode off
        # src: on-source
        # ref: off-source
        #######################

        # second to day
        t_src = t_src/60/60/24
        t_ref = t_ref/60/60/24
        n_repeat = n_repeat
        t_change = t_change/60/60/24

        mjd_src_start = self.mjd_a[0] + np.arange(n_repeat)*(t_src + t_change + t_ref + t_change)
        mjd_ref_start = mjd_src_start + (t_src + t_change)

        is_ = self.in_range(self.mjd_on, mjd_src_start, t_src)
        self.p_on_src = self.p_on[is_]
        self.inds_on_src = self.inds_on[is_]

        is_ = self.in_range(self.mjd_off, mjd_src_start, t_src)
        self.p_off_src = self.p_off[is_]
        self.inds_off_src = self.inds_off[is_]

        is_ = self.in_range(self.mjd_on, mjd_ref_start, t_ref)
        self.p_on_ref = self.p_on[is_]
        self.inds_on_ref = self.inds_on[is_]

        is_ = self.in_range(self.mjd_off, mjd_ref_start, t_ref)
        self.p_off_ref = self.p_off[is_]
        self.inds_off_ref = self.inds_off[is_]

        self.mjd_src_start = mjd_src_start
        self.mjd_ref_start = mjd_ref_start

    def plot_sep(self,):
        try:
            if getattr(self, 'plot', False):
                figname = self.out_name_base + "-sep.pdf"
                plot_sep(self.inds_on, self.inds_off, self.p_on, self.p_off, figname=figname)
        except:
            pass

    def gen_Ta(self, only_off=False):
        """
        using

        self.p_on_src
        self.p_off_src
        self.p_on_ref
        self.p_off_ref

        to gen

        self.Ta as shape (1,chan,polar)

        """

        p_cal_src = np.mean(self.p_on_src, axis=0, dtype='float64', keepdims=True) - \
                    np.mean(self.p_off_src, axis=0, dtype='float64', keepdims=True)
        p_cal_ref = np.mean(self.p_on_ref, axis=0, dtype='float64', keepdims=True) - \
                    np.mean(self.p_off_ref, axis=0, dtype='float64', keepdims=True)
        # smooth power of cal
        p_cal_src_s = self._get_smoothed(p_cal_src)
        p_cal_ref_s = self._get_smoothed(p_cal_ref)

        Ta_on_src = (np.mean(self.p_on_src, axis=0, dtype='float64', keepdims=True)/p_cal_src_s - 1)*self.Tcal_s
        Ta_off_src = (np.mean(self.p_off_src, axis=0, dtype='float64', keepdims=True)/p_cal_src_s)*self.Tcal_s
        Ta_on_ref = (np.mean(self.p_on_ref, axis=0, dtype='float64', keepdims=True)/p_cal_ref_s - 1)*self.Tcal_s
        Ta_off_ref = (np.mean(self.p_off_ref, axis=0, dtype='float64', keepdims=True)/p_cal_ref_s)*self.Tcal_s

        # use interaged time as weight
        if not only_off:
            Ta = np.average(np.r_[Ta_on_src, Ta_off_src], axis=0, weights=[len(self.inds_on_src),len(self.inds_off_src)]) - \
                 np.average(np.r_[Ta_on_ref, Ta_off_ref], axis=0, weights=[len(self.inds_on_ref),len(self.inds_off_ref)])
        else:
            Ta = np.average(Ta_off_src, axis=0) - \
                 np.average(Ta_off_ref, axis=0)


        Ta = Ta[None,]

        self.Ta = Ta

        # also change mjd
        inds_tmp = self.inds_off_src[:1]
        self.mjd = self.mjd_a[inds_tmp]

    def gen_radec(self,):
        """
        try to calculation for each record: self.ra_a, self.dec_a
        use the position of first src record as self.ra, self.dec
        also set self.mjd and self.mjd_a

        """

        from .radec import get_radec
        nB = self.nB
        guess_str = self.fname_part

        radec = get_radec(self.mjd_a, guess_str=guess_str, nBs=[nB,])
        self.ra_a, self.dec_a = radec[f'ra{nB}'], radec[f'dec{nB}']

        inds_tmp = self.inds_off_src[:1]
        self.ra, self.dec = self.ra_a[inds_tmp], self.dec_a[inds_tmp]
        self.mjd = self.mjd_a[inds_tmp]

    def plot_radec(self, figsize=[10,10], outname=None):
        """
        plot the ra dec to check the separation of src and ref
        """

        from matplotlib import pyplot as plt
        plt.figure(figsize=figsize)
        plt.scatter(self.ra_a[self.inds_on_src], self.dec_a[self.inds_on_src], s=1, color='r')
        plt.scatter(self.ra_a[self.inds_off_src], self.dec_a[self.inds_off_src], s=1, color='r', label='src')
        plt.scatter(self.ra_a[self.inds_on_ref], self.dec_a[self.inds_on_ref], s=1, color='b')
        plt.scatter(self.ra_a[self.inds_off_ref], self.dec_a[self.inds_off_ref], s=1, color='b', label='ref')
        plt.xlabel('ra')
        plt.ylabel('dec')
        plt.grid()
        plt.minorticks_on()
        plt.legend()
        if outname is not None:
            print(f'Saving ra dec plot to {outname}')
            plt.savefig(outname)





__all__ = ['get_vcorr', 'freq2vel', 'vel2freq', 'frame_correct_freq', 'spectrum_resample', 'correct_spec']


import numpy as np
import scipy.interpolate as interp

import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord, EarthLocation
from astropy.coordinates import ICRS, LSR, LSRK, LSRD

from . import conf


def get_vcorr(ra, dec, mjd, frame='BARYCENT'):
    """

    Parameters
    ----------
    ra, dec: scalar or ndarray
             unit in deg
    mjd: scalar or ndarray

    frame: str
        'BARYCENT', 'HELIOCEN', 'LSR', 'LSRK', 'LSRD'

    Returns
    -------
    vcorr: scalar or ndarray
           unit in km/s
    """

    if frame == 'BARYCENT' or 'LSR' in frame:
        kind = 'barycentric'
    elif frame == 'HELIOCEN':
        kind = 'heliocentric'
    else:
        raise(ValueError(f'frame {frame} not support'))
    # telescope location from conf
    geog = EarthLocation.from_geodetic(lat=conf.lat*u.rad, lon=conf.long*u.rad, height=conf.height*u.m)
    c = SkyCoord(ra=ra*u.deg, dec=dec*u.deg, frame="icrs")
    vcorr = c.radial_velocity_correction(kind=kind, obstime=Time(mjd, format='mjd'), location=geog)
    if frame in ['BARYCENT', 'HELIOCEN']:
        return vcorr.to(u.km/u.s).value

    # to Local standard of rest
    frame_dict = {'LSR': LSR,
              'LSRK': LSRK,
              'LSRD': LSRD,
             }

    if not np.isscalar(ra):
        pm_ra_cosdec = np.full_like(ra, 0)*u.mas/u.yr # must be zero
        pm_dec = np.full_like(ra, 0)*u.mas/u.yr # must be zero
        distance = np.full_like(ra, 1)*u.pc # arbitrary value
    else:
        pm_ra_cosdec = 0*u.mas/u.yr
        pm_dec = 0*u.mas/u.yr
        distance = 1*u.pc
    vel = ICRS(ra=ra*u.deg, dec=dec*u.deg, pm_ra_cosdec=pm_ra_cosdec, pm_dec=pm_dec, radial_velocity=vcorr, distance=distance)
    vcorr = vel.transform_to(frame_dict[frame]()).radial_velocity

    return vcorr.to(u.km/u.s).value



def freq2vel(freq, vtype='radio'):
    """
    freq to velocity

    Parameters
    ----------
    freq: array
         shape: (n,); unit in Mhz
    vtype: str
          'radio' or 'optical', default is 'radio'
    """
    restfreq = conf.restfreq
    if vtype == 'radio':
        vel = (restfreq - freq)/restfreq*conf.vlight
    elif vtype == 'optical':
        vel = (restfreq - freq)/freq*conf.vlight
    return vel

def vel2freq(vel, vtype='radio'):
    """
    velocity to freq

    Parameters
    ----------
    vel: array
         shape: (n,); unit in km/s
    vtype: str
          'radio' or 'optical', default is 'radio'
    """
    restfreq = conf.restfreq
    if vtype == 'radio':
        freq = restfreq * (1 - vel / conf.vlight)
    elif vtype == 'optical':
        freq = restfreq / (vel / conf.vlight + 1)
    return freq

def _freq_vcorr(freq, vcorr):
    """
    doppler correct ``freq`` (Mhz) with ``vcorr`` (km/s)
    """
    return freq * np.sqrt((conf.vlight - vcorr)/(conf.vlight + vcorr))

def frame_correct_freq(freq, ra, dec, mjd, frame='BARYCENT'):
    """
    dopper correct the ``freq`` of the spectra at (ra, dec) observed at time ``mjd``

    freq : array like (m,)
          unit in Mhz
    mjd, ra, dec: scalar or array (n,)
    frame: str; 'BARYCENT', 'HELIOCEN', 'LSR', 'LSRK', 'LSRD'

    return: shape (m,) if mjd, ra, dec are scalar; shape (n, m) if array
    """
    isscalar = False
    if np.isscalar(mjd):
        mjd = np.array([mjd])
        isscalar = True
    if np.isscalar(ra):
        ra = np.array([ra])
        isscalar = True
    if np.isscalar(dec):
        dec = np.array([dec])
        isscalar = True

    vcorr = get_vcorr(ra, dec, mjd, frame)
    if not isscalar:
        vcorr = vcorr[:, None]
    return _freq_vcorr(freq, vcorr)



def spectrum_resample(freq, flux, freq_new, method='interp', interp_kind='linear'):
    """
    resample spectrum with ``flux`` at ``freq`` to `freq_new`
    """
    if method == 'interp':
        flux_new = interp.interp1d(freq, flux, kind=interp_kind, bounds_error=False, fill_value=np.nan)(freq_new)
        return flux_new

def correct_spec(Ta, freq, ra, dec, mjd, frame='LSRK', method='interp', interp_kind='linear'):
    """
    doppler correct spectra with ``Ta`` at ``freq`` and resample the spectra at same frequency sample points

    Parameters
    ----------
    Ta: array; shape: (m,n,2) or (m,n)
    freq: array; shape: (n,) in ascending order
    ra, dec: array; shape: (m,)
    frame: str
          rest frame, 'BARYCENT', 'HELIOCEN', 'LSR', 'LSRK', 'LSRD'
    """
    if Ta.ndim == 2:
        Ta = Ta[:,:,None]
        ndim_ori = 2
    elif Ta.ndim == 3:
        ndim_ori = 3
    else:
        raise(ValueError('input ndim 2 or 3'))

    # velocity of observer (telescope) with respect to frame
    vcorr = get_vcorr(ra, dec, mjd, frame)

    # new freq sample points
    f_d = freq[1] - freq[0]
    fcorr = _freq_vcorr(freq, np.median(vcorr))
    tmp = np.hstack([np.arange(freq[0]-f_d, fcorr.min(),-f_d)[::-1], freq, np.arange(freq[-1]+f_d, fcorr.max(),f_d)])
    freq_new = tmp[(tmp> fcorr.min()) & (tmp< fcorr.max())]

    Ta_new = np.zeros((Ta.shape[0], len(freq_new), Ta.shape[2]))
    for i in range(Ta.shape[0]):
        for j in range(Ta.shape[2]):
            fcorr_i = _freq_vcorr(freq, vcorr[i])
            Ta_new[i,:,j] = spectrum_resample(fcorr_i, Ta[i,:,j], freq_new, method=method, interp_kind=interp_kind)
    if ndim_ori == 2:
        Ta_new = Ta_new[..., 0]
    return Ta_new, freq_new





__all__ = ['CustomStdout', 'set_output']


import os
import sys
from .colors import Colors as c


class CustomStdout:
    def __init__(self, original_stdout, additional_info=''):
        self.original_stdout = original_stdout
        self.additional_info = additional_info
        self.additional_info = f'{c.bold}{self.additional_info}{c.end}'
        self.lineends = True

    def _write_with_additional_info(self, message):
        if message.strip() in ['\n', ''] or message.startswith('File exists') or message.startswith('Saved to'):
            self.original_stdout.write(message)
        else:
            self.original_stdout.write(f'{self.additional_info}{message}')

    def write(self, message):
        # tqdm start with '\r' but not with '\n'
        if message.startswith('\r'):
            message = message[1:]  # Properly handle messages that intend to overwrite the line
            self.original_stdout.write(f'\r{self.additional_info}{message}')
            # or not add additional info?
            self.lineends = True
        else:
            if not message.endswith('\n'):
                if self.lineends:
                    self._write_with_additional_info(message)
                else:
                    self.original_stdout.write(message)
                self.lineends = False
            else:
                if self.lineends:
                    self._write_with_additional_info(message)
                else:
                    self.original_stdout.write(message)
                self.lineends = True


    def flush(self):
        self.original_stdout.flush()

    def __getattr__(self, attr):
        """ Delegate access to other attributes and methods to the original stdout """
        return getattr(self.original_stdout, attr)

def set_output(additional_info):
    # if HIFAST_MAKE_OUTPUT_CLEAR is set in environment variable and 1
    if 'HIFAST_MAKE_OUTPUT_CLEAR' in os.environ and os.environ['HIFAST_MAKE_OUTPUT_CLEAR'] == '1':
        from tqdm import tqdm
        from functools import partialmethod
        # after add the additional to the progress bar in tqdm, the linewith will exceed the terminal width, so we need to set nclos to 0, a small number
        # or import shutil;shutil.get_terminal_size().columns - len(additional_info) - extra_to_make_sure_it_will_not_exceed_the_terminal_width
        tqdm.__init__ = partialmethod(tqdm.__init__, ncols=0, mininterval=5)

        sys.stdout = CustomStdout(sys.stdout, additional_info)
        sys.stderr = CustomStdout(sys.stderr, additional_info)



from .utils.io import *

sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}", formatter_class=formatter_class, allow_abbrev=False,
                        description='replace fields in fpath_ori with that from fpath_new', )
parser.add_argument('fpath_ori',
                    help='file path')
parser.add_argument('fpath_new',
                    help='file path')
parser.add_argument('--fields', nargs='+', required=True,
                    help='fields will be replaced, e.g. mjd')
parser.add_argument('-y', '--yes', action='store_true', not_in_write_out_config_file=True,
                    help="Do not ask for confirmation.")


if __name__ == '__main__':
    import h5py
    args_ = parser.parse_args()
    # print(parser.format_help())
    # print("----------")
    print('#'*35+'Args'+'#'*35)
    print(parser.format_values())  # useful for logging where different settings came from
    print('#'*35+'####'+'#'*35)
    args = args_

    f_ori =  h5py.File(args.fpath_ori, 'r+')
    S_ori = f_ori['S']
    f_new =  h5py.File(args.fpath_new, 'r')
    S_new = f_new['S']

    print(f"""replace {','.join(args.fields)} in
{args.fpath_ori}
with that in
{args.fpath_new}
""")
    if not args.yes:
        QA = input('input y[es] to continue\n')
        if not (QA.lower() == 'y' or QA.lower() == 'yes'):
            print('aborting...')
            f_ori.close()
            f_new.close()
            sys.exit(0)
    for key in args.fields:
        print(f"replacing {key}")
        S_ori[key][...] = S_new[key][:]

    his = rec_his()
    for key in his.keys():
        if 'HISTORY' in key:
            pass
            f_ori['Header'].attrs[key] = his[key]

    f_ori.close()
    f_new.close()



#!/usr/bin/env python
# coding: utf-8
# Author: Ziming liu zmliu@nao.cas.cn
# Yingjie Jing
import numpy as np
import scipy.interpolate as interp



from astropy.coordinates import SkyCoord, EarthLocation
from astropy import coordinates as coord
from astropy import units as u
from astropy.time import Time

import os

import json

from . import conf

def Gain_para():
    fpath = os.path.dirname(__file__)+'/data/FAST_gain_curve.txt'
    # Manually parse the file to replicate:
    # gain_para.set_index([0,1],inplace=True)
    # gain_para = gain_para[gain_para.columns[::2]] (taking every 2nd column?)
    # original code: gain_para= pd.read_csv(..., header=None, sep='\s+')
    # gain_para.set_index([0,1]) -> first two cols are index: Beam, Param
    # gain_para= gain_para[gain_para.columns[::2]] -> if original cols were 0,1,2,3,4... after set_index, cols are 2,3,4...
    # The ::2 on columns likely means skipping error columns or similar.
    # gain_para.columns = list(range(1050, 1500, 50))
    
    # Structure needed: data[beam][freq][param] or data[beam][param][freq]
    # Existing usage: gain_para.loc[f'M{nB:02d}'][fre][['a','b','c']]
    # This implies: result[beam][freq] -> gives a series/dict where keys are 'a','b','c'
    # Wait, gain_para.loc['M01'] gives a DF indexed by 'a','b','c'.
    # gain_para.loc['M01'][fre] gives the column at 'fre', which is a Series with index 'a','b','c'.
    # So we need: data[beam][freq] = {'a': val, 'b': val, 'c': val}
    
    data = {}
    with open(fpath, 'r') as f:
        for line in f:
            parts = line.strip().split()
            if not parts: continue
            beam = parts[0]
            param = parts[1]
            values = parts[2:]
            
            # Original: gain_para = gain_para[gain_para.columns[::2]]
            # If values indices are 0,1,2,3... ::2 takes 0, 2, 4...
            selected_values = values[::2]
            
            freqs = list(range(1050, 1500, 50))
            if len(selected_values) != len(freqs):
                 # Fallback or error? Assuming file matches expected length
                 pass
            
            if beam not in data: data[beam] = {}
            
            for i, freq in enumerate(freqs):
                if freq not in data[beam]: data[beam][freq] = {}
                data[beam][freq][param] = float(selected_values[i])
                
    # To support syntax: gain_para.loc['M01'][fre]['a']
    # We can wrap it in a class or just return a dict-like object.
    # But existing code: gain_para.loc[f'M{nB:02d}'][fre][['a','b','c']]
    # If data is dict: data['M01'][fre] returns {'a':..., 'b':..., 'c':...}
    # Dicts don't support list indexing [['a','b','c']].
    # We need to change the usage in `ZA2gain` as well.
    return data 

def Get_ZA(ra, dec, mjd):
    """
    ra: deg
    dec: deg
    mjd: day
    -------------------
    return ZA: deg
    """
    obs_location = EarthLocation.from_geodetic(lat=conf.lat*u.rad, lon=conf.long*u.rad, height=conf.height*u.m)
    aa_frame =  coord.AltAz(obstime = Time(mjd,format='mjd'), location=obs_location)

    crd= SkyCoord(ra, dec,unit='deg',)
    crd_aa = crd.transform_to(aa_frame)
    za =90 - crd_aa.alt.deg
    return za

def ZA2gain(ZA, nB):
    """
    nB: Beam number
    ZA: zenith angle; deg
    """
    # get eta
    gain_para= Gain_para()
    # freq_key= np.array(gain_para.columns) 
    # Use keys from first available beam (e.g. M01)
    # The keys are frequencies.
    # gain_para structure: data[beam][freq][param]
    first_beam = list(gain_para.keys())[0]
    freq_key = np.array(list(gain_para[first_beam].keys()))
    
    gain= np.zeros((len(ZA),len(freq_key)))
    is_use= ZA > 26.4
    for i,fre in enumerate(freq_key):
        # a,b,c= gain_para.loc[f'M{nB:02d}'][fre][['a','b','c']]
        params = gain_para[f'M{nB:02d}'][fre]
        a, b, c = params['a'], params['b'], params['c']
        
        gain[is_use,i]= c * ZA[is_use] + b + 26.4*(a - c)
        gain[~is_use,i]= a * ZA[~is_use] + b
    # 25.6*eta
    gain *= 25.6
    return gain, freq_key

def Get_gain(ra, dec, mjd, nB, freq=None):
    """
    ra: deg
    dec: deg
    mjd: day
    nB: int
    freq: None or array
    ------------------------
    return gain
           freq_key
    """
    ZA= Get_ZA(ra, dec, mjd)
    gain,freq_key=ZA2gain(ZA, nB)
    if freq is not None:
        gain= interp.interp1d(freq_key,gain,kind='quadratic',fill_value='extrapolate')(freq)
        freq_key=freq
    return gain, freq_key


def gain_diff_from_ZA(ZA_1: float, ZA_2: np.array, nB: int, freq: np.array) -> np.array:
    """
    Calculate the gain difference between two different ZA's (zenith angle)

    :param ZA_1: scalar value for first ZA
    :param ZA_2: 1d array values for second ZA
    :param nB: scalar integer 
    :param freq: 1d array values for frequency
   
    :return: array of gain differences for each frequency. (ZA_2 - ZA_1)
    """
    # eta diff
    
    fpath = os.path.dirname(__file__) + '/data/gain_ZA_fit.json'
    
    # Open and load the json data into a dictionary
    with open(fpath,'r') as f:
        coeffs_dict = json.load(f)

    fun_a = np.poly1d(coeffs_dict['a'][f'M{nB:02}'])
    fun_c = np.poly1d(coeffs_dict['c'][f'M{nB:02}'])

    # Calculate the delta between two ZA's
    delta = (ZA_2 - ZA_1)[:, None]
    diffs = np.zeros((len(delta), len(freq)), dtype=float)
    
    # Check for ZA_2 angles below 26.4
    is_s = ZA_2 <= 26.4

    # Compute diffs based on the whether the ZA's are below or above 26.4
    if ZA_1 <= 26.4:
        diffs[is_s] = delta[is_s] * fun_a(freq)
        diffs[~is_s] = ((ZA_2[~is_s] - 26.4)[:, None] * fun_c(freq) +
                         (26.4 - ZA_1) * fun_a(freq))
    elif ZA_1 > 26.4:
        diffs[~is_s] = delta[~is_s] * fun_c(freq)
        diffs[is_s] = ((ZA_2[is_s] - 26.4)[:, None] * fun_a(freq) +
                         (26.4 - ZA_1) * fun_c(freq))
    #  25.6*eta
    diffs *= 25.6
    return diffs


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/I__bld_i.ipynb (unless otherwise specified).

__all__ = ['w_conf', 'Test', 'phrase_ylim', 'main']

# Cell
import numpy as np
import h5py
import matplotlib.pyplot as plt
import mpl_interactions.ipyplot as iplt
import ipywidgets as widgets
from ipywidgets import interact

import re
import os
from glob import glob

from .core.baseline import get_baseline, sub_baseline
from .utils.misc import extend_Trues
from .utils.io import PolarMjdChan_to_MjdChanPolar

from scipy import ndimage

import warnings
warnings.filterwarnings("ignore", r'overflow encountered in exp')

# Cell
w_conf = {}
w_conf['continuous_update'] = False
w_conf['orientation'] = 'horizontal'
w_conf['readout'] = True
w_conf['layout'] = {}


def _IntSlider(readout_format='d', **kwargs):
    """
    kwargs:
    slider = (value=7, min=0, max=10, step=1)
    """
    res = {}
    for key in kwargs.keys():
        res[key] = widgets.IntSlider(*kwargs[key], description=f'{key}',
                                     **w_conf, readout_format=readout_format)
    return res


def _BoundedIntText(readout_format='d', **kwargs):
    """
    kwargs:
    slider = (value=7, min=0, max=10, step=1)
    """
    res = {}
    for key in kwargs.keys():
        res[key] = widgets.BoundedIntText(*kwargs[key], description=f'{key}',
                                          **w_conf, readout_format=readout_format)
    return res


def _IntText(readout_format='d', **kwargs):
    """
    kwargs:
    slider = (value=7, min=0, max=10, step=1)
    """
    res = {}
    for key in kwargs.keys():
        res[key] = widgets.IntText(*kwargs[key], description=f'{key}',
                                   **w_conf, readout_format=readout_format)
    return res


def _FloatSlider(readout_format='.2f', **kwargs):
    """
    kwargs:
    slider = (value=7, min=0, max=10, step=1)
    """
    res = {}
    for key in kwargs.keys():
        v = kwargs[key]
        res[key] = widgets.FloatSlider(value=v[0], min=v[1], max=v[2], step=v[3], description=f'{key}',
                                       **w_conf, readout_format=readout_format)
    return res


def _FloatLog10Slider(readout_format='.2f', **kwargs):
    """
    log10
    kwargs:
    slider = (value, min exponent of base, max exponent of base, exponent step)
    """
    res = {}
    for key in kwargs.keys():
        v = kwargs[key]
        res[key] = widgets.FloatLogSlider(value=v[0], base=10, min=v[1], max=v[2], step=v[3], description=f'{key}',
                                          **w_conf, readout_format=readout_format)
    return res


def _RadioButtons(**kwargs):
    """
    kwargs:
    slider = options
    """
    res = {}
    for key in kwargs.keys():
        v = kwargs[key]
        res[key] = widgets.RadioButtons(options=v, value=v[0], description=f'{key}',
                                        **w_conf,)
    return res
# widgets.RadioButtons(
#     options=['pepperoni', 'pineapple', 'anchovies'],
# #    value='pineapple', # Defaults to 'pineapple'
# #    layout={'width': 'max-content'}, # If the items' names are long


def _Dropdown(**kwargs):
    """
    kwargs:
    slider = options
    """
    res = {}
    for key in kwargs.keys():
        v = kwargs[key]
        res[key] = widgets.Dropdown(options=v, value=v[0], description=f'{key}',
                                    **w_conf,)
    return res

# Cell
class Test(object):
    def __init__(self, T2p, freq, frange=None):

        self.T2p = T2p
        freq = freq[:]
        if frange is not None:
            self.is_use = (freq > frange[0]) & (freq < frange[1])
            self.freq = freq[self.is_use]
        else:
            self.is_use = None
            self.freq = freq
        self.select()

    def select(self, start=0, length=20, polar=0):
        self.T2p_t = PolarMjdChan_to_MjdChanPolar(self.T2p[polar:polar+1, start:start+length])
        if self.is_use is not None:
            self.T2p_t = self.T2p_t[:, self.is_use]
    def sub(self, x, start=0, length=20, polar=0, **kwargs):
        self.select(start, length, polar)
        self.bld = sub_baseline(self.freq, self.T2p_t,
                                verbose=False, **kwargs)
        return np.full(x.shape, np.nan)

    def get_ori(self, x, i, **kwargs):
        res = self.T2p_t[i, :, 0]
        return np.vstack([res, ndimage.gaussian_filter1d(res, 3)]).T

    def get_bld(self, x, i, **kwargs):
        res = self.bld[i, :, 0]
        return np.vstack([res, ndimage.gaussian_filter1d(res, 3)]).T

    def get_bl(self, x, i,  **kwargs):
        return self.T2p_t[i, :, 0] - self.bld[i, :, 0]

    def get_bld_mean(self, x, start_stop,  **kwargs):
        start, stop = start_stop
        res = np.mean(self.bld[int(start):int(stop)+1, :, 0], axis=0)
        return np.vstack([res, ndimage.gaussian_filter1d(res, 3)]).T

# Cell
def phrase_ylim(ylim, vals):

    ylim = list(map(lambda x: float(x) if 'per' != str(x)[:3] else
                    np.nanpercentile(vals, float(str(x)[3:]), interpolation='nearest'), ylim))
    return ylim

# Cell
def main():
    global ylim

    tes = Test(T2p, freq, frange)

    sliders = {}
    sliders.update(_BoundedIntText(
        start=(0, 0, tes.T2p.shape[1]-1-length, 1), polar=(0, 0, 1, 1)))
    sliders.update(_IntSlider(njoin=(0, 1, length, 1)))
    sliders.update(_Dropdown(method=('arPLS', 'asPLS', 'masPLS', 'poly', 'Chebyshev'),
                             s_method_freq=('gaussian', 'boxcar', 'none'),
                             s_method_t=('none', 'gaussian', 'boxcar'),
                             ))
    sliders.update(_IntSlider(s_sigma_freq=(3, 1, 20, 1), s_sigma_t=(3, 1, 20, 1)))
    sliders.update(_FloatLog10Slider(lam=(1e8, 5, 13, 0.2), readout_format='.2e'))
    sliders.update(_IntSlider(deg=(2, 1, 10, 1),
                              niter=(100, 1, 200, 1)))



    plt.ioff()
    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=figsize)

    if isinstance(ylim, str):
        if ylim == 'auto':
            ylim = phrase_ylim(['per0.1', 'per99.5'], tes.T2p_t)
        elif ylim == 'full':
            ylim = 'fixed'
    else:
        ylim = phrase_ylim(ylim, tes.T2p_t)

    if isinstance(ylim, str):
        ylim2 = ylim
    else:
        ylim2 = [(ylim[0]-ylim[1])/2, (ylim[1]-ylim[0])/2]
    xlim = [tes.freq.min(), tes.freq.max()]

    # bld parameter
    # slider_formats = {'lam': "{:.2e}",
    #                   'njoin': "{:d}",
    #                   }

    controls = iplt.plot(tes.freq, tes.sub, nproc=nproc,
                         **sliders,
                         color='r', linewidth=2, xlim=xlim, ylim=ylim, ax=ax1, play_buttons=True, display_controls=False)

    controls2 = iplt.plot(tes.freq, tes.get_ori, i=range(length),
                          controls=controls, xlim=xlim, ylim=ylim, ax=ax1, display_controls=False, play_buttons=True)
    controls3 = iplt.plot(tes.freq, tes.get_bl,
                          controls=controls2, xlim=xlim, ylim=ylim, ax=ax1)

    iplt.plot(tes.freq, tes.get_bld,
              controls=controls2, xlim=xlim, ylim=ylim2, ax=ax2)
    plt.axhline(y=0,)

    controls4 = iplt.plot(tes.freq, tes.get_bld_mean, start_stop=('r', 0, length, 11),
                          controls=controls, xlim=xlim, ylim=ylim2, ax=ax3)

    plt.axhline(y=0, )
    ax1.set_title('origial spectra', fontsize=8)
    ax2.set_title('spectra after baseline removed', fontsize=8)

    fig.tight_layout()

    w = controls.controls
    hbs = [
        widgets.HBox([widgets.Label(value=f"1. Select spectra from $start$(max:{controls.controls['start'].max}) to $start+{length}$ to fit:"),
                      w['start'], w['polar']]),
        widgets.HBox([widgets.Label(value=f"2. Fitting:"), w['njoin'], ]),
        widgets.HBox([w['s_method_t'], w['s_sigma_t'],
                     w['s_method_freq'], w['s_sigma_freq']]),
        widgets.HBox([w['method'], w['lam'], w['deg'], w['niter']]),
        widgets.HBox(
            [widgets.Label(value=f"3. Show one spectra in top and middle panels:"), w['i']]),
        widgets.HBox([widgets.Label(
            value=f"4. Show the stack spectra of in the range in the bottom panel:"), w['start_stop']]),
    ]

    BOX = widgets.VBox(hbs)
    display(BOX)
    plt.show()
    display(BOX)



#!/usr/bin/env python
# coding: utf-8

# Chuan-Peng Zhang cpzhang@nao.cas.cn
# Yingjie Jing

import numpy as np
import scipy.interpolate as interp
from scipy import ndimage

from ..utils.misc import average_every_n, boxcar_smooth1d, mask_Trues, extend_Trues, smooth1d
from scipy import ndimage

def get_mean_rms(freq, T2p, **kwargs):
    rfi_nan = (freq>1390.0) & (freq<1410.0)
    m = np.mean(T2p[:, rfi_nan], axis=1)[:,None]
    std= np.std(T2p[:, rfi_nan], axis=1)[:,None]
    return m, std

def get_rfi_folder(freq, T2p, folder, times, ext_add=0, ext_frac=0):
    rms_times = times
    nth = np.arange(len(T2p))
    T2p_m = average_every_n(T2p, folder, axis=0, drop=False)
    nth_m = average_every_n(nth, folder, axis=-1,drop=False)

    m, std = get_mean_rms(freq, T2p_m)
    is_rfi = (T2p_m > m+std*rms_times) | (T2p_m < m-std*rms_times)
    
    if ext_add > 0 or ext_frac > 0:
        is_rfi = extend_Trues(is_rfi, axis=1, ext_add=ext_add, ext_frac=ext_frac)
    
    inds = interp.interp1d(nth_m, range(len(nth_m)), kind='nearest',fill_value='extrapolate')(nth).astype('int')

    return is_rfi[inds]


def find_signal(T2p, *, s_method_t='gaussian', s_sigma_t=10, s_method_freq='gaussian', s_sigma_freq=5, 
              times_s=2, times=5, ext_add=0, ext_frac=0):
    """
    T2p: array, shape=(x,N,2); T or flux with two polar and sorted by time
    s_sigma: int; gaussian smooth size along time (axis=0)
    chan_smooth_method: method of smoothing along channel axis
    times_s: # for smoothed channel
    times: # for not smoothed channel
    """
    # smooth along time
    if s_method_t is not None: 
        T2p_s = smooth1d(T2p, s_method_t, s_sigma_t, axis=0)
    else:
        T2p_s = T2p

    # smooth channel
    if s_method_freq is not None:
        T2p_sc = smooth1d(T2p_s, s_method_freq, s_sigma_freq, axis=1)
    else:
        T2p_sc = T2p_s

    # estimated rms
    pers = np.nanpercentile(T2p_s, [10, 50, 90], axis=1)
    rms = np.min(abs(np.diff(pers,axis=0)),axis=0)
    # find rfi
    # compare channel smoothed 
    is_exceeded = abs(T2p_sc) >= times_s*rms[:,None]
    # compare origin 
    is_exceeded = is_exceeded | (abs(T2p_s) >= times*rms[:,None])
    
    is_exceeded = is_exceeded[:,:,0] | is_exceeded[:,:,1]
    if ext_add > 0 or ext_frac > 0:
        is_exceeded = extend_Trues(is_exceeded, axis=1, ext_add=ext_add, ext_frac=ext_frac)
    #return is_exceeded
    # n continue along t axis as rfi
    return is_exceeded

def get_rfi_c(T2p, n_continue=50, return_sig=False, **kwargs):
    is_sig = find_signal(T2p, **kwargs)
    is_rfi = mask_Trues(is_sig, axis=0, leng_lim=n_continue)
    if return_sig:
        return is_rfi, is_sig
    return is_rfi

def mask_rfi_t(freq, T2p, method='smooth', **kwargs):
    if method=='smooth':
        return get_rfi_c(T2p, **kwargs)
    elif method=='folder':
        return get_rfi_folder(freq, T2p, **kwargs)
    return is_rfi

## compare "signal" with rfi
def cross_rfi_1d(arr, is_rfi, frac_match=0.3):
    """
    is_rfi:
    frac_match:
    """
    leng_lim = 1
    # add False in the beginning and ending of arr
    arr = np.hstack([[False],arr,[False]])
    diff = np.diff(arr.astype('int16'))

    ind_neg = np.where(diff==-1)[0]
    ind_posi = np.where(diff==1)[0]

    leng = (ind_neg - ind_posi)
    is_use = leng > leng_lim 

    for i,j,_leng in zip(ind_posi[is_use],ind_neg[is_use], leng[is_use]):
        leng_in = np.sum(is_rfi[i:j])
        if leng_in/_leng < frac_match:
            arr[i+1:j+1] = False
    return arr[1:-1] | is_rfi

def cross_rfi_axis1_d2(arr, is_rfi, inplace=False, **kwarg):
    if not inplace:
        out = np.zeros_like(is_rfi)
    else:
        out = is_rfi
    for i in range(len(arr)):
        out[i] = cross_rfi_1d(arr[i], is_rfi[i], **kwarg)
    return out





__all__ = []


from .utils.io import *


sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                        formatter_class=formatter_class, allow_abbrev=False,
                        description='Convert file to new format', )
add_common_argument(parser)
parser.add_argument('fpath',
                    help='input hdf5 file path.')


if __name__ == '__main__':
    args = parser.parse_args()
    dict_in = load_hdf5_to_dict_old(args.fpath)

    if 'T' in dict_in.keys():
        wcs_data_name = 'T'
    elif 'Ta' in dict_in.keys():
        wcs_data_name = 'Ta'
    elif 'flux' in dict_in.keys():
        wcs_data_name = 'flux'
    else:
        raise(ValueError('can not find spec'))

    if 'frame' in dict_in['Header'].keys():
        if dict_in[wcs_data_name].ndim == 2:
            dict_in[wcs_data_name] = dict_in[wcs_data_name][..., None]

    dict_in[wcs_data_name] = MjdChanPolar_to_PolarMjdChan(dict_in[wcs_data_name])

    save_specs_hdf5(os.path.join(args.outdir, os.path.basename(args.fpath)), dict_in)



import os
import sys
import time
import subprocess
import argparse
import shlex
from glob import glob


# from importlib import import_module

# subcomands_type1 = [
#     'sep',
#     'pos_swi',
#     'radec',
#     'bld',
#     'flux',
#     'sw',
#     'rfi',
#     'multi',
#     'add_radec',
#     'sub_ref',
#     'convert',
#     'downsample',
#      ]

# subcomands_type2 = [
#     
#     'waterfall',
#      ]

subcomands_excl = ['cube', 'cube2', 'waterfall', 'funcs', 'find']
subcomands_type1 = []
for fpath in glob(os.path.dirname(__file__) + '/[a-z,A-Z]*.py'):
    subcomand = os.path.basename(fpath)[:-3]
    if subcomand not in subcomands_excl:
        subcomands_type1 += [subcomand, ]
subcomands_type1.sort()
subcomands_type2 = []



## parser

parser = argparse.ArgumentParser(prog=f"python -m hifast", formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False, add_help=False,
                        description='', usage="python -m hifast subcomand fpath1 [fpath2 fpath3 ...] ... [-p P] [Args with '-'] ... [Args with '--'] ...")
parser.add_argument('-h', '--help', action='store_true',
                   help='')
parser.add_argument('--version', action='store_true',
                   help='display version')
subparsers = parser.add_subparsers(help='', dest="subcomand", description='')

for subcomand in subcomands_type1:
    parser_a = subparsers.add_parser(subcomand, formatter_class=argparse.RawDescriptionHelpFormatter,
                 add_help=False,
                  )
    parser_a.add_argument('fpath', nargs="*", help='input files path: fpath1 [fpath2 fpath3 ...]')
    parser_a.add_argument('-p', type=int, default=1,
                    help='max processes at a time')
    parser_a.add_argument('-h', '--help', action='store_true',
                    help='')
for subcomand in subcomands_type2:
    parser_a = subparsers.add_parser(subcomand, formatter_class=argparse.RawDescriptionHelpFormatter,
                 add_help=False,
                                )
    parser_a.add_argument('-h', '--help', action='store_true',
                   help='')

## RUN
args, remain = parser.parse_known_args()

if args.version:
    from .__init__ import __version__
    print(__version__)
    sys.exit(0)


if args.subcomand is None:
    print(parser.format_help())
    sys.exit(0)
elif args.help:
    subcomand = args.subcomand
    if subcomand in subcomands_type1:
        _help = f"""
usage: python -m hifast {subcomand} fpath1 [fpath2 fpath3 ...] [-p P] [Arguments of hifast.{subcomand} with '-' or '--' ]

Type1: RUN ``python -m hifast.{subcomand}`` for files in parallel with ``P`` processes.

optional arguments:
-p P        max processes at a time
-h, --help

Below is the help from hifast.{subcomand}:
        """
    elif subcomand in subcomands_type2:
        _help = f"""
Same with hifast.{subcomand}.

Below is the help from hifast.{subcomand}:
        """
    command = f"python -m hifast.{args.subcomand} -h"
    print(_help)
    sys.stdout.flush()
    # help_sub = subprocess.check_output(command, shell=True)
    # help_sub = help_sub.decode('utf-8')
    # print(_help, help_sub)
    # sys.exit(0)
elif args.subcomand in subcomands_type1:
    fpaths = '\n'.join(args.fpath)
    command = f"printf '{fpaths}' | xargs -I {{}} -n 1 -P {args.p} {sys.executable} -m hifast.{args.subcomand} {{}} {shlex.join(remain)}"
elif args.subcomand in subcomands_type2:
    command = f"{sys.executable} -m hifast.{args.subcomand} {shlex.join(remain)}"

p = subprocess.Popen(
        command,
        stderr=sys.stderr, stdout=sys.stdout,
        shell=True
          )
try:
    p.wait()
except KeyboardInterrupt:
    p.kill()





__all__ = ['ky_dir_default', 'plot_radec', 'guess_ky', 'process_ky', 'kydata2radec', 'radec_interp', 'get_radec']


import os
import warnings
import re
import sys
from glob import glob
import json

import numpy as np
import scipy.interpolate as interp

import h5py

from astropy import units as u
from astropy.time import Time
from astropy.utils import iers
#iers.Conf.iers_auto_url.set("https://datacenter.iers.org/data/9/finals2000A.all")
from . import kypara2radec
from ..utils.io import rec_his, save_specs_hdf5
import erfa


ky_dir_default = [os.path.expanduser("~")+'/KY/',
                    '/data/inspur_disk06/fast_data/KY/',
                    '/data31/KY/',]


def _tight_ra(ra):
    if len(ra)<=1:
        return ra
    ra_s= np.sort(ra)
    diff_s= np.diff(ra_s)
    ind_max= np.argmax(diff_s)
    if diff_s[ind_max] < ra_s[0]+360-ra_s[-1]:
        return ra
    else:
        ra= np.copy(ra)
        is_c= ra>=ra_s[ind_max+1]
        ra[is_c]= ra[is_c]-360
        return ra


def plot_radec(radec,outname=None,ax=None):
    if ax is None:
        from matplotlib import pyplot as plt
        if outname is not None:
            plt.switch_backend('agg')
        fig, ax = plt.subplots(1,1,figsize=(8,8))
    ind_sort = np.argsort(radec['mjd'])
    for i in range(1,20):
        ax.plot(_tight_ra(radec[f'ra{i}'][ind_sort]), radec[f'dec{i}'][ind_sort])

    ax.set_xlabel('RA')
    ax.set_ylabel('DEC')
    ax.minorticks_on()
    if outname is not None:
        ax.set_title(outname)
        fig.tight_layout()
        fig.savefig(outname)


def guess_ky(fname, obs_mjd, ky_dir=None):
    """
    fname: string
    obs_mjd: float; mjd
    """
    obs_data= (Time(obs_mjd,format='mjd')+8*u.hour).to_datetime().strftime('%Y_%m_%d')

    if ky_dir is None:
        for ky_dir in ky_dir_default:
            print(f'Check if KY directory {ky_dir} exists...')
            if os.path.exists(ky_dir):
                print(f'Search KY .xlsx file in {ky_dir}')
                break
            else:
                print(f'KY directory {ky_dir} does not exist.')
    if ky_dir is None:
        raise(ValueError('No KY directory found'))
    keys= re.match('^.*M[0-1][0-9]_[W,N,F]',os.path.basename(fname)).group(0)[:-6].split('_')
    ky_files=[] #set variable first, in case len(keys)==1
    for i in range(1,len(keys)):
        ky_files= glob(f"{ky_dir}/**/{'_'.join(keys[:-i])}*{obs_data}*.xlsx", recursive=True)
        if len(ky_files)>=1:
            break
    return ky_files


def process_ky(ky_files, mjds_src=None, tol=0, ky_fixed=False):
    '''
    processing KY files, return ky_data, mjd, ky_file

    ky_files: list; path of feed (馈源舱) recording file (.xlsx)
    mjds_src: array; if not None, will check if the feed file cover the mjds_src
    tol: float; tolerable time, unit is second
    '''
    tol = float(tol)/24/3600 # second to day
    for ky_file in ky_files:
        print(f"check {ky_file}")
        this = True
        utcoffset = 8*u.hour
        #read ky data
        sheet_name = '整控-馈源舱数据'
        import openpyxl
        wb = openpyxl.load_workbook(ky_file, data_only=True, read_only=True)
        # Select sheet
        if sheet_name in wb.sheetnames:
             ws = wb[sheet_name]
        else:
             print("can't find '整控-馈源舱数据', try to use the first sheet")
             ws = wb.active

        # Read headers
        rows = ws.rows
        headers = [cell.value for cell in next(rows)]
        # Filter None headers
        headers = [h if h else f"Unnamed:{i}" for i, h in enumerate(headers)] # Handle empty headers if any? 
        # But ky_data usually has nice headers.

        data_dict = {h: [] for h in headers}
        
        count = 0
        for row in rows:
             count += 1
             for i, cell in enumerate(row):
                 if i < len(headers):
                     data_dict[headers[i]].append(cell.value)
                     
        if count == 0:
            print('empty sheet')
            continue
            
        ky_data = data_dict
        systime = np.array(ky_data['SysTime'], dtype=str)
        systime = Time(systime, format='iso', scale='utc') - utcoffset # covert local time to UTC
        mjd = systime.mjd
        ky_data['systime_utc'] = systime
        if mjds_src is None:
            # if only deal with ky file
            return ky_data, mjd, ky_file

        diff = mjds_src.min()-mjd.min()
        if -tol < diff < 0:
            if not ky_fixed:
                print(f"The earliest input mjd is earlier by {abs(diff)*24*3600:.3f}s than the mjd in KY. Which is in the set tolerable time, RA and Dec will be extrapolated.")
            else:
                print(f"The earliest input mjd is earlier by {abs(diff)*24*3600:.3f}s than the mjd in KY.")
        if -tol > diff:
            this = False
            print(f"The earliest input mjd is earlier by {abs(diff)*24*3600:.3f}s than the mjd in KY.")
            #raise(ValueError(f"input mjd exceed {abs(diff)*24*3600}s to the mjd in KY,"))
        if ky_fixed:
            diff = mjd.max()-mjds_src.min()
            if -tol < diff< 0:
                print(f"the begin of input mjd is later {abs(diff)*24*3600:.3f}s than the end of mjd in KY. ")
            if -tol > diff:
                this = False
                print(f"the begin of input mjd is later {abs(diff)*24*3600:.3f}s than the end of mjd in KY. Abort.")
                #raise(ValueError(f"input mjd exceed {abs(diff)*24*3600}s to the mjd in KY,"))

        diff = mjd.max() - mjds_src.max()
        if -tol < diff< 0:
            if not ky_fixed:
                print(f"The last input mjd is later by {abs(diff)*24*3600:.3f}s than the mjd in KY. Which is in the set tolerable time, RA and Dec will be extrapolated.")
            else:
                print(f"The last input mjd is later by {abs(diff)*24*3600:.3f}s than the mjd in KY.")
        if -tol > diff:
            #raise(ValueError(f"input mjd exceed {abs(diff)*24*3600}s to the mjd in KY,"))
            if not ky_fixed:
                this = False
                print(f"The last input mjd is later by {abs(diff)*24*3600:.3f}s than the mjd in KY. Abort.")

        if this:
            break

    if not this:
        raise(ValueError(f"The input mjd are not in the mjd range of those KY files"))
    else:
        print(f"Using {ky_file}")
        return ky_data, mjd, ky_file


def kydata2radec(ky_data, mjd, nBs='All', ky_fixed=False, nproc=1, backend='astropy'):
    """
    calculating ra dec of ky from ky data
    ky_data: data in feed file (.xlsx)
    mjd: mjd of ky_data if ky_fixed is False, else mjd of the source
    nBs: list or str:'All'
    """
    if nBs== 'All':
        nBs = range(1,20)

    multibeamAngle= np.asarray(ky_data['SDP_AngleM'], dtype='float64')

    #实测中心波束相对中心的全局坐标
    globalCenterX = np.asarray(ky_data['SDP_PhaPos_X'], dtype='float64')
    globalCenterY = np.asarray(ky_data['SDP_PhaPos_Y'], dtype='float64')
    globalCenterZ = np.asarray(ky_data['SDP_PhaPos_Z'], dtype='float64')
    #实测下平台的全局姿态角
    globalYaw = np.asarray(ky_data['SDP_SwtDPose_Y'], dtype='float64')
    globalPitch = np.asarray(ky_data['SDP_SwtDPose_P'], dtype='float64')
    globalRoll = np.asarray(ky_data['SDP_SwtDPose_R'], dtype='float64')

    if ky_fixed:
        warnings.warn("By using the ky_fixed parameter, the feed is assumed to be stationary during observation.")
        std= np.std(multibeamAngle[-10:])/np.pi*180
        if std > 3:
            raise(ValueError(f'multibeamAngle variation at last 10 points is {std} degree.'))
        full_fun= lambda x:np.full(len(mjd), np.mean(x[-5:]))
        multibeamAngle= full_fun(multibeamAngle)
        globalCenterX = full_fun(globalCenterX)
        globalCenterY = full_fun(globalCenterY)
        globalCenterZ = full_fun(globalCenterZ)
        globalYaw =     full_fun(globalYaw)
        globalPitch =   full_fun(globalPitch)
        globalRoll =    full_fun(globalRoll)

    radec= {}
    radec['mjd']= mjd
    radec['angle'] = multibeamAngle
    #ymdhms = ky_data['systime_utc'].ymdhms
    #utc1, utc2 = erfa.dtf2d(b"UTC", ymdhms['year'], ymdhms['month'], ymdhms['day'], ymdhms['hour'], ymdhms['minute'], ymdhms['second'])
    if ky_fixed:
        obstime = Time(mjd, format='mjd', scale='utc')
    else:
        obstime = ky_data['systime_utc']

    if nproc > 1:
        import multiprocessing
        global fun_mp
        def fun_mp(para):
            nB, verbose = para
            print(f'beam {nB}')
            return kypara2radec.kypara2radec(obstime, multibeamAngle, nB,
                                              globalCenterX,  globalCenterY, globalCenterZ, globalYaw, globalPitch, globalRoll,
                                               backend=backend, verbose=verbose)
        with multiprocessing.Pool(processes=nproc) as pool:
            res = pool.map(fun_mp, zip(nBs, [True,] + [False, ]*(len(nBs)-1)))
        for i, nB in enumerate(nBs):
            radec['ra'+str(nB)]= res[i][0]
            radec['dec'+str(nB)]= res[i][1]
            radec['Az'+str(nB)]= res[i][2]
            radec['ZD'+str(nB)]= res[i][3]
    else:
        for ii, nB in enumerate(nBs):
            verbose = True if ii == 0 else False
            ra_, dec_, Az_, ZD_ = kypara2radec.kypara2radec(obstime, multibeamAngle, nB,
                                          globalCenterX,  globalCenterY, globalCenterZ, globalYaw, globalPitch, globalRoll, backend=backend, verbose=verbose)
            radec['ra'+str(nB)]= ra_
            radec['dec'+str(nB)]= dec_
            radec['Az'+str(nB)]= Az_
            radec['ZD'+str(nB)]= ZD_
            print(f'beam {nB}')
    return radec


def radec_interp(radec_ky, mjds_src):
    """
    get ra dec of source through interpolating that of ky
    """
    mjds_ky= radec_ky['mjd']
    radec_src={}
    radec_src['mjd']= mjds_src
    print('interpolating...')
    radec_src['is_extrapo']= (mjds_src < np.nanmin(mjds_ky)) | (mjds_src > np.nanmax(mjds_ky))
    for key in radec_ky:
        if 'ra' in key or 'dec' in key or 'angle' in key or 'Az' in key or 'ZD' in key:
            value = radec_ky[key]
            if 'ra' in key:
                value = _tight_ra(value)
            radec_src[key]= interp.interp1d(mjds_ky, value, kind='linear', fill_value ='extrapolate')(mjds_src)
            if 'ra' in key:
                radec_src[key][radec_src[key]<0] += 360 # PyAstronomy.pyasl don't support negative ra
    return radec_src


def get_radec(parse_mjd, guess_str=None, ky_files=None, tol=0.3, ky_fixed=False, use_cache=True, nproc=1,
              backend='astropy', env_para=None, dUT1=None, nBs='All', cache_kyradec=True):
    """
    parse_mjd: str or array
               str: *.hdf5 or *.xlsx
               array: shape is (m,); need input guess_str or ky_files
    guess_str: str; used to guess feed file name, for example, "M31_Drift_v3_6_arcdrift-M19_W"
               if None, use parse_mjd instead and parse_mjd must be str.
    ky_files: list of str; specify feed files; if None, will guess from guess_str or parse_mjd (str).
    tol: float; unit is second
    ky_fixed: bool; if True, assuming the feed is stationary during observation.
    use_cache: bool; use cached radec of feed
    nproc: int; parallel cpu number
    backend: str; erfa or astropy
    env_para: dict; e.g.
            env_para = {}
            env_para['phpa'] = 925.
            env_para['temperature'] = 15
            env_para['humidity'] = 0.8
    dUT1: UT1 - UTC
    """
    if nBs== 'All':
        nBs = range(1,20)
    elif isinstance(nBs, list):
        use_cache = False
        cache_kyradec = False
    else:
        raise(ValueError("nBs need be 'All' or a list with numbers"))

    ident = {}
    if env_para is not None:
        kypara2radec.set_obs_env_para(env_para['phpa'], env_para['temperature'], env_para['humidity'])
        ident.update(env_para)
    if dUT1 is not None:
        kypara2radec.dUT1 = dUT1
        if backend == 'erfa':
            ident['dUT1'] = dUT1
    import json
    from hashlib import md5
    ident_str = md5(json.dumps(ident).encode('utf-8')).hexdigest()

    # if input feed file, return ra dec in it
    if isinstance(parse_mjd, str):
        fname = parse_mjd
        if 'xls' in fname.split('.')[-1]:
            ky_data, mjd, ky_file = process_ky([fname,])
            return kydata2radec(ky_data, mjd, nBs=nBs, ky_fixed=False, backend=backend, nproc=nproc)
        elif 'hdf5' in fname.split('.')[-1]:
            with h5py.File(fname,'r') as f:
                    mjds_src = f['S']['mjd'][:]
        else:
            raise(ValueError('fname needs end with xlsx or hdf5'))
    else:
        mjds_src = parse_mjd
        if guess_str is None and ky_files is None:
            raise(ValueError('need input `guess_str` or `ky_files`'))
    if guess_str is None and isinstance(parse_mjd, str):
        guess_str = parse_mjd


    # obtain cache_file and cache_key
    cache_dir= os.path.expanduser('~/.cache/fast_python/kyradec/')
    os.makedirs(cache_dir, exist_ok=True)
    cache_file = cache_dir + 'known_kyfile.json'
    obs_data = (Time(mjds_src[0], format='mjd')+8*u.hour).to_datetime().strftime('%Y_%m_%d')
    try:
        cache_key = re.match('^.*M[0-1][0-9]_[W,N,F]', os.path.basename(guess_str)).group(0)[:-6] + '-' + obs_data
    except:
        cache_key = None
        use_cache = False

    # find feed file corresponding to guess_str
    ky_file = None
    if ky_files is None:
        # try to find feed file corresponding to guess_str
        if use_cache:
            #if cached file is not exits, init
            if not os.path.exists(cache_file):
                    with open(cache_file, "w") as f:
                        json.dump({}, f)
            try:
                with open(cache_file,'r') as f:
                    ky_file= json.load(f)[cache_key]
                print('get kyfile name from ', cache_file)
                #print('using ', ky_file)
                ky_data = None
            except KeyError:
                pass
    # if can't get ky_file from cache
    if ky_file is None:
        if ky_files is None:
            obs_mjd = mjds_src[0]
            ky_files = guess_ky(guess_str, obs_mjd)
            if len(ky_files) == 0:
                raise(ValueError("can not find KY file similar to input file, please appoint KY files" ))
    else:
        ky_files = [ky_file,]
    # test kyfiles and load data
    res = process_ky(ky_files, mjds_src, tol, ky_fixed=ky_fixed)
    if res is None:
        return
    else:
        ky_data, mjds_ky, ky_file = res
    # cache ky_file
    if cache_key is not None and cache_kyradec:
        import tempfile
        file_tmp = tempfile.mktemp(prefix=os.path.basename(cache_file)+'.tmp.', dir=os.path.dirname(cache_file))
        with open(cache_file, "r") as f:
            cache = json.load(f)
        cache.update({cache_key:ky_file})
        with open(file_tmp,'w') as f:
            json.dump(cache, f, indent=4)
            f.flush()
            os.fsync(f.fileno())
        os.rename(file_tmp, cache_file)
    # obtain radec of ky
    cache_fname_radec = os.path.join(cache_dir,'.'.join(os.path.basename(ky_file).split('.')[:-1]) + f'_{backend}_cache_{ident_str}.npz')
    cache_fname_radec_b = os.path.join(cache_dir,'.'.join(os.path.basename(ky_file).split('.')[:-1]) + f'_{backend}_cache_{ident_str}.npy')
    radec_ky = None
    if use_cache and not ky_fixed:
        try:
            radec_ky = dict(np.load(cache_fname_radec).items())
            print('using cached ra dec of KY: ', cache_fname_radec)
        except:
            try:
                radec_ky = np.load(cache_fname_radec_b, allow_pickle=True)[()]
                print('using cached ra dec of KY: ', cache_fname_radec_b)
            except:
                pass
    if radec_ky is None:
        if ky_data is None:
            # ky_file is obtained from cache_file
            ky_data, mjds_ky, ky_file= process_ky([ky_file,], mjds_src, tol, ky_fixed=ky_fixed)
        if ky_fixed:
            return kydata2radec(ky_data, mjds_src, ky_fixed=ky_fixed, nproc=nproc, backend=backend, nBs=nBs)
        else:
            radec_ky= kydata2radec(ky_data, mjds_ky, ky_fixed=ky_fixed, nproc=nproc, backend=backend, nBs=nBs)
            if cache_kyradec:
                print('saving radec of KY to cached file')
                np.savez(cache_fname_radec, **radec_ky)
    return radec_interp(radec_ky, mjds_src)




"""
hifast.multi
============

Multi-step operations for spectral data processing.

This module performs several key processing steps on the input spectra:
1.  **Reference Frame Correction**: Converts velocities from the telescope's Topocentric frame to a standard frame (e.g., LSRK).
2.  **RFI Masking**: Replaces flagged RFI data with NaN.
3.  **Polarization Merging**: Averages XX and YY polarizations.
4.  **Extrapolation Masking**: Optionally masks data points with extrapolated coordinates.
"""

__all__ = ['IO', 'parser']


from .utils.io import *


def create_parser():
    sep_line = '##'+'#'*70+'##'
    parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}", formatter_class=formatter_class, allow_abbrev=False,
                            description='Perform multi-step operations: RFI masking, reference frame correction, and polarization merging.')
    add_common_argument(parser)

    # --- Input/Output ---
    group = parser.add_argument_group('Input/Output')
    group.add_argument('fpath', metavar='FILE',
                        help='Input spectra file path (HDF5 format).')
    group.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')], metavar=('MIN', 'MAX'),
                        help='Limit the frequency range to process [MHz]. Followed by two values (min, max).')

    # --- Frame Correction ---
    group = parser.add_argument_group('Frame Correction')
    group.add_argument('--fc', type=bool_fun, choices=[True, False], default='True', not_in_write_out_config_file=True,
                       help='Enable reference frame correction (Topocentric -> LSRK/Heliocentric).')
    group.add_argument('--frame', choices=['BARYCENT', 'HELIOCEN', 'LSRK', 'LSRD'], default='LSRK',
                       help='Target reference frame. "LSRK" (Local Standard of Rest, Kinematic) is commonly used.')
    group.add_argument('--vtype', choices=['radio', 'optical'], default='optical',
                       help='Velocity definition type. "optical" (cz) or "radio". Note: This is primarily for tracking/single-spectrum analysis. `hifast.cube` calculates velocity from frequency and ignores this column.')

    # --- Data Processing ---
    group = parser.add_argument_group('Data Processing')
    group.add_argument('--replace_rfi', type=bool_fun, choices=[True, False], default='True',
                       help='Replace RFI-contaminated spectra with NaN (requires "is_rfi" dataset in input).')
    group.add_argument('--mask_extrapo', type=bool_fun, choices=[True, False], default='False',
                       help='Mask data points where RA/DEC coordinates were extrapolated (requires "is_extrapo" dataset).')
    group.add_argument('--merge_polar', type=bool_fun, choices=[True, False], default='True',
                       help='Merge (average) the two polarizations (XX and YY) into a single Stokes I intensity.')
    
    return parser

parser = create_parser()


class IO(BaseIO):
    ver = 'old'
    def _get_fpart(self,):
        """
        need modify this function
        """
        fpart = '-fc'
        return fpart

    def __call__(self, save=True):
        import numpy as np
        args = self.args
        s2p = self.s2p[:]
        # try load is_rfi
        if 'is_rfi' in self.fs.keys():
                # if not replace_rfi, frame correct it

                if self.is_use_freq is not None:
                    inds = np.where(self.is_use_freq)[0]
                    is_rfi = self.fs['is_rfi'][:, inds]
                else:
                    is_rfi = self.fs['is_rfi'][:]
                if args.replace_rfi:
                    print('replacing RFI')
                    s2p[is_rfi] = np.nan
        else:
            is_rfi = None

        # mask data points where ra/dec are extrapolated
        if args.mask_extrapo and 'is_extrapo' in self.fs.keys():
            is_extrapo = self.fs['is_extrapo'][:]
            print('masking data points with extrapolated ra/dec')
            # is_extrapo shape: (Mjd,), s2p shape: (Mjd, Chan, Polar)
            s2p[is_extrapo, :, :] = np.nan

        if args.merge_polar and s2p.ndim == 3:
            print('average two polarization...')
            s2p = np.mean(s2p, axis=2, keepdims=True)
        # frame
        if args.fc:
            from .core.corr_vel import freq2vel, correct_spec
            if 'frame' in self.Header.keys():
                raise(ValueError(f'rest frame already corrected... please set --fc as False'))
                sys.exit()
            else:
                self.Header['frame'] = args.frame
                self.Header['vel_type'] = args.vtype
            # correct is_rfi
            if is_rfi is not None and not args.replace_rfi:
                is_rfi, _ = correct_spec(is_rfi, self.freq, self.ra, self.dec, self.mjd,
                                           frame=args.frame, method='interp', interp_kind='nearest')
                is_rfi = np.array(is_rfi, dtype=bool)
            print('frame correcting...')
            s2p, freq = correct_spec(s2p, self.freq, self.ra, self.dec, self.mjd, frame=args.frame, method='interp')
            vel = freq2vel(freq, vtype=args.vtype)
            self.s2p_out = s2p
            self.gen_dict_out(freq=freq, vel=vel)  # replace freq, add vel
            if 'Tcal' in self.dict_out.keys():
                # use ``try`` to be compatible with the file generated with old version having bug
                try:
                    values = self.dict_out['Tcal'] # self.is_use_freq has been applied; still is (Mjd,Chan,Polar)
                    self.dict_out['Tcal'], _ = correct_spec(values, self.freq, self.ra, self.dec, self.mjd, frame=args.frame, method='interp')
                except:
                    pass
        else:
            self.s2p_out = s2p
            self.gen_dict_out()
        # save is_rfi only if replace_rfi is False
        if not args.replace_rfi and is_rfi is not None:
            self.dict_out['is_rfi'] = is_rfi
        else:
            if 'is_rfi' in self.dict_out.keys():
                self.dict_out.pop('is_rfi')
        if save:
            self.save()


if __name__ == '__main__':
    args_ = parser.parse_args()
    # print(parser.format_help())
    # print("----------")
    print('#'*35+'Args'+'#'*35)
    args_from = parser.format_values()
    print(args_from)
    print('#'*35+'####'+'#'*35)

    HistoryAdd = {'args_from': args_from} if args_.my_config is not None else None
    io = IO(args_, HistoryAdd=HistoryAdd)
    io()





__all__ = ['IO']


from .utils.io import *
import copy


sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}", formatter_class=formatter_class, allow_abbrev=False,
                        description='', )
add_common_argument(parser)
parser.add_argument('fpath',
                    help='input spectra temperature file path.')
parser.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')],
                    help='Limit frequence range')
parser.add_argument('--irange', type=int, nargs=2,
                    help="")


class IO(BaseIO):
    ver = 'new'

    def _get_fpart(self,):
        """
        need modify this function
        """
        fpart = '-sli'
        return fpart

    def __call__(self, save=True):
        import numpy as np

        args = self.args
        type1 = ['ra', 'dec',  'is_delay', 'is_extrapo', 'is_on', 'mjd', 'next_to_cal']
        type2 = ['freq', 'vel']

        dict_out = {}
        dict_out["Header"] = self.Header
        for key in self.fs.keys():
            if key in type1:
                dict_out[key] = self.fs[key][args.irange[0]:args.irange[1]]
            elif key in type2:
                if self.is_use_freq is not None:
                    inds = np.where(self.is_use_freq)[0]
                    dict_out[key] = self.fs[key][inds[0]:inds[-1]+1]
                else:
                    dict_out[key] = self.fs[key]
            elif key == 'Tcal':
                if self.is_use_freq is not None:
                    inds = np.where(self.is_use_freq)[0]
                    dict_out[key] = self.fs[key][:, inds[0]:inds[-1]+1]
                else:
                    dict_out[key] = self.fs[key]
            elif key in ['is_excluded', 'is_rfi']:
                if self.is_use_freq is not None:
                    inds = np.where(self.is_use_freq)[0]
                    dict_out[key] = self.fs[key][args.irange[0]:args.irange[1], inds[0]:inds[-1]+1]
                else:
                    dict_out[key] = self.fs[key][args.irange[0]:args.irange[1]]
            elif key == self.infield:
                dict_out[key] = self.s2p[:, args.irange[0]:args.irange[1]]

        self.dict_out = dict_out
        if save:
            self.save()


if __name__ == '__main__':

    args_ = parser.parse_args()
    # print(parser.format_help())
    # print("----------")
    print('#'*35+'Args'+'#'*35)
    args_from = parser.format_values()
    print(args_from)
    print('#'*35+'####'+'#'*35)

    HistoryAdd = {'args_from': args_from} if args_.my_config is not None else None
    args_.no_radec = True
    io = IO(args_, HistoryAdd=HistoryAdd)
    io()
